"""
Author: Wenyu Ouyang
Date: 2023-04-06 14:45:34
LastEditTime: 2024-04-10 20:23:36
LastEditors: Wenyu Ouyang
Description: Test the multioutput model
FilePath: \torchhydro\tests\test_train_camelspro_mtl.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

import numpy as np
import torch

from torchhydro.configs.config import update_cfg
from torchhydro.models.crits import MultiOutLoss, RMSELoss
from torchhydro.trainers.trainer import train_and_evaluate


def test_cuda_available():
    assert torch.cuda.is_available()


def test_multiout_loss_nan_gap():
    data0 = torch.tensor([2.0]).repeat(8, 3, 1)
    data1 = torch.tensor(np.full((8, 3, 1), np.nan))
    data1[0, 0, :] = 1.0
    data1[3, 0, :] = 2.0
    data1[6, 0, :] = 3.0
    data1[1, 1, :] = 4.0
    data1[4, 1, :] = 5.0
    data1[7, 1, :] = 6.0
    data1[2, 2, :] = 7.0
    data1[5, 2, :] = 8.0
    targ = torch.cat((data0, data1), dim=-1)
    pred = torch.tensor(np.full((8, 3, 2), 1.0))
    # device = torch.device('cpu')
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    targ = targ.to(device)
    pred = pred.to(device)
    rmse = RMSELoss()
    r = MultiOutLoss(rmse, data_gap=[0, 1], device=[0], item_weight=[1, 1])
    # for sum, we ignore last interval
    expect_value = rmse(
        torch.tensor(np.array([1.0, 2.0, 4.0, 5.0, 7.0]).astype(float)),
        torch.tensor([3.0, 3.0, 3.0, 3.0, 3.0]),
    ) + rmse(data0, torch.tensor(np.full((8, 3, 1), 1.0)))
    np.testing.assert_almost_equal(
        r(pred, targ).cpu().detach().numpy(), expect_value.cpu().detach().numpy()
    )


def test_flow_et_mtl(mtl_args, config_data):
    """
    Test for data augmentation of flow -> et with KuaiLSTM

    Parameters
    ----------
    config_data

    Returns
    -------

    """
    update_cfg(config_data, mtl_args)
    train_and_evaluate(config_data)



"""
The method is similar with dpl4xaj.py.
The difference between dpl4xaj and dpl4xaj_nn4et is:
in the former, the parameter of PBM is only one output of a DL model,
while in the latter, time series output of a DL model can be as parameter of PBM 
and some modules could be replaced with neural networks
"""

from typing import Union
import torch
from torch import nn
from torch import Tensor
from torch.nn import functional as F

from torchhydro.configs.model_config import MODEL_PARAM_DICT, MODEL_PARAM_TEST_WAY
from torchhydro.models.ann import SimpleAnn
from torchhydro.models.kernel_conv import KernelConv
from torchhydro.models.simple_lstm import SimpleLSTM
from torchhydro.models.dpl4xaj import (
    calculate_prcp_runoff,
    linear_reservoir,
    xaj_sources,
    xaj_sources5mm,
)

PRECISION = 1e-5


class NnModule4Hydro(nn.Module):
    """A NN module for Hydrological model.
    Generally, the difference between it and normal NN is:
    we need constrain its output to some specific value range

    Parameters
    ----------
    nn : _type_
        _description_
    """

    def __init__(
        self,
        nx: int,
        ny: int,
        hidden_size: Union[int, tuple, list] = None,
        dr: Union[float, tuple, list] = 0.0,
    ):
        """
        A simple multi-layer NN model with final linear layer

        Parameters
        ----------
        nx
            number of input neurons
        ny
            number of output neurons
        hidden_size
            a list/tuple which contains number of neurons in each hidden layer;
            if int, only one hidden layer except for hidden_size=0
        dr
            dropout rate of layers, default is 0.0 which means no dropout;
            here we set number of dropout layers to (number of nn layers - 1)
        """
        super(NnModule4Hydro, self).__init__()
        self.ann = SimpleAnn(nx, ny, hidden_size, dr)

    def forward(self, x, w0, prcp, pet, k):
        """the forward function of the NN ET module

        Parameters
        ----------
        x : _type_
            _description_
        w0 : _type_
            water storage
        p : _type_
            precipitation
        pet: tensor
            potential evapotranspiration, used to be part of upper limit of ET
        k: tensor
            coefficient of PET in XAJ model, used to be part of upper limit of ET

        Returns
        -------
        _type_
            _description_
        """
        zeros = torch.full_like(w0, 0.0, device=x.device)
        et = torch.full_like(w0, 0.0, device=x.device)
        w_mask = w0 + prcp > PRECISION
        y = self.ann(x)
        z = y.flatten()
        et[w_mask] = torch.clamp(
            z[w_mask],
            min=zeros[w_mask],
            # torch.minimum computes the element-wise minimum: https://pytorch.org/docs/stable/generated/torch.minimum.html
            # k * pet is real pet in XAJ model
            max=torch.minimum(w0[w_mask] + prcp[w_mask], k[w_mask] * pet[w_mask]),
        )
        return et


def calculate_1layer_w_storage(um, lm, dm, w0, pe, r):
    """
    Update the soil moisture value.

    According to the runoff-generation equation 2.60 in the book "SHUIWENYUBAO", dW = dPE - dR

    Parameters
    ----------
    um
        average soil moisture storage capacity of the upper layer (mm)
    lm
        average soil moisture storage capacity of the lower layer (mm)
    dm
        average soil moisture storage capacity of the deep layer (mm)
    w0
        initial values of soil moisture
    pe
        net precipitation; it is able to be negative value in this function
    r
        runoff

    Returns
    -------
    torch.Tensor
        w -- soil moisture
    """
    xaj_device = pe.device
    tensor_zeros = torch.full_like(w0, 0.0, device=xaj_device)
    # water balance (equation 2.2 in Page 13, also shown in Page 23)
    w = w0 + pe - r
    return torch.clamp(w, min=tensor_zeros, max=(um + lm + dm) - PRECISION)


class Xaj4DplWithNnModule(nn.Module):
    """
    XAJ model for Differential Parameter learning with neural network as submodule
    """

    def __init__(
        self,
        kernel_size: int,
        warmup_length: int,
        nn_module=None,
        param_var_index=None,
        source_book="HF",
        source_type="sources",
        et_output=1,
        nn_hidden_size: Union[int, tuple, list] = None,
        nn_dropout=0.2,
        param_test_way=MODEL_PARAM_TEST_WAY["time_varying"],
    ):
        """
        Parameters
        ----------
        kernel_size
            the time length of unit hydrograph
        warmup_length
            the length of warmup periods;
            XAJ needs a warmup period to generate reasonable initial state values
        nn_module
            We initialize the module when we firstly initialize Xaj4DplWithNnModule.
            Then we will iterately call Xaj4DplWithNnModule module for warmup.
            Hence, in warmup period, we don't need to initialize it again
        param_var_index
            the index of parameters which will be time-varying
            NOTE: at the most, we support k, b, and c to be time-varying
        et_output
            we only support one-layer et now, because its water balance is not easy to handle with
        """
        if param_var_index is None:
            param_var_index = [0, 6]
        if nn_hidden_size is None:
            nn_hidden_size = [16, 8]
        super(Xaj4DplWithNnModule, self).__init__()
        self.params_names = MODEL_PARAM_DICT["xaj_mz"]["param_name"]
        param_range = MODEL_PARAM_DICT["xaj_mz"]["param_range"]
        self.k_scale = param_range["K"]
        self.b_scale = param_range["B"]
        self.im_sacle = param_range["IM"]
        self.um_scale = param_range["UM"]
        self.lm_scale = param_range["LM"]
        self.dm_scale = param_range["DM"]
        self.c_scale = param_range["C"]
        self.sm_scale = param_range["SM"]
        self.ex_scale = param_range["EX"]
        self.ki_scale = param_range["KI"]
        self.kg_scale = param_range["KG"]
        self.a_scale = param_range["A"]
        self.theta_scale = param_range["THETA"]
        self.ci_scale = param_range["CI"]
        self.cg_scale = param_range["CG"]
        self.kernel_size = kernel_size
        self.warmup_length = warmup_length
        # there are 2 input variables in XAJ: P and PET
        self.feature_size = 2
        if nn_module is None:
            # 7: k, um, lm, dm, c, prcp, p_and_e[:, 1] + 1/3: w0 or wu0, wl0, wd0
            self.evap_nn_module = NnModule4Hydro(
                7 + et_output, et_output, nn_hidden_size, nn_dropout
            )
        else:
            self.evap_nn_module = nn_module
        self.source_book = source_book
        self.source_type = source_type
        self.et_output = et_output
        self.param_var_index = param_var_index
        self.nn_hidden_size = nn_hidden_size
        self.nn_dropout = nn_dropout
        self.param_test_way = param_test_way

    def xaj_generation_with_new_module(
        self,
        p_and_e: Tensor,
        k,
        b,
        im,
        um,
        lm,
        dm,
        c,
        *args,
        # wu0: Tensor = None,
        # wl0: Tensor = None,
        # wd0: Tensor = None,
    ) -> tuple:
        # make sure physical variables' value ranges are correct
        prcp = torch.clamp(p_and_e[:, 0], min=0.0)
        pet = torch.clamp(p_and_e[:, 1], min=0.0)
        # wm
        wm = um + lm + dm
        if self.et_output != 1:
            raise NotImplementedError("We only support one-layer evaporation now")
        w0_ = args[0]
        if w0_ is None:
            w0_ = 0.6 * (um.detach() + lm.detach() + dm.detach())
        w0 = torch.clamp(w0_, max=wm - PRECISION)
        concat_input = torch.stack([k, um, lm, dm, c, w0, prcp, pet], dim=1)
        e = self.evap_nn_module(concat_input, w0, prcp, pet, k)
        # Calculate the runoff generated by net precipitation
        prcp_difference = prcp - e
        pe = torch.clamp(prcp_difference, min=0.0)
        r, rim = calculate_prcp_runoff(b, im, wm, w0, pe)
        if self.et_output == 1:
            w = calculate_1layer_w_storage(
                um,
                lm,
                dm,
                w0,
                prcp_difference,
                r,
            )
            return (r, rim, e, pe), (w,)
        else:
            raise ValueError("et_output should be 1")

    def forward(self, p_and_e, parameters_ts, return_state=False):
        """
        run XAJ model

        Parameters
        ----------
        p_and_e
            precipitation and potential evapotranspiration
        parameters_ts
            time series parameters of XAJ model;
            some parameters may be time-varying specified by param_var_index
        return_state
            if True, return state values, mainly for warmup periods

        Returns
        -------
        torch.Tensor
            streamflow got by XAJ
        """
        xaj_device = p_and_e.device
        if self.param_test_way == MODEL_PARAM_TEST_WAY["time_varying"]:
            parameters = parameters_ts[-1, :, :]
        else:
            # parameters_ts must be a 2-d tensor: (basin, param)
            parameters = parameters_ts
        # denormalize the parameters to general range
        # TODO: now the specific parameters are hard coded; 0 is k, 1 is b, 6 is c, same as in model_config.py
        if 0 not in self.param_var_index or self.param_var_index is None:
            ks = self.k_scale[0] + parameters[:, 0] * (
                self.k_scale[1] - self.k_scale[0]
            )
        else:
            ks = self.k_scale[0] + parameters_ts[:, :, 0] * (
                self.k_scale[1] - self.k_scale[0]
            )
        if 1 not in self.param_var_index or self.param_var_index is None:
            bs = self.b_scale[0] + parameters[:, 1] * (
                self.b_scale[1] - self.b_scale[0]
            )
        else:
            bs = self.b_scale[0] + parameters_ts[:, :, 1] * (
                self.b_scale[1] - self.b_scale[0]
            )
        im = self.im_sacle[0] + parameters[:, 2] * (self.im_sacle[1] - self.im_sacle[0])
        um = self.um_scale[0] + parameters[:, 3] * (self.um_scale[1] - self.um_scale[0])
        lm = self.lm_scale[0] + parameters[:, 4] * (self.lm_scale[1] - self.lm_scale[0])
        dm = self.dm_scale[0] + parameters[:, 5] * (self.dm_scale[1] - self.dm_scale[0])
        if 6 not in self.param_var_index or self.param_var_index is None:
            cs = self.c_scale[0] + parameters[:, 6] * (
                self.c_scale[1] - self.c_scale[0]
            )
        else:
            cs = self.c_scale[0] + parameters_ts[:, :, 6] * (
                self.c_scale[1] - self.c_scale[0]
            )
        sm = self.sm_scale[0] + parameters[:, 7] * (self.sm_scale[1] - self.sm_scale[0])
        ex = self.ex_scale[0] + parameters[:, 8] * (self.ex_scale[1] - self.ex_scale[0])
        ki_ = self.ki_scale[0] + parameters[:, 9] * (
            self.ki_scale[1] - self.ki_scale[0]
        )
        kg_ = self.kg_scale[0] + parameters[:, 10] * (
            self.kg_scale[1] - self.kg_scale[0]
        )
        # ki+kg should be smaller than 1; if not, we scale them
        ki = torch.where(
            ki_ + kg_ < 1.0,
            ki_,
            (1 - PRECISION) / (ki_ + kg_) * ki_,
        )
        kg = torch.where(
            ki_ + kg_ < 1.0,
            kg_,
            (1 - PRECISION) / (ki_ + kg_) * kg_,
        )
        a = self.a_scale[0] + parameters[:, 11] * (self.a_scale[1] - self.a_scale[0])
        theta = self.theta_scale[0] + parameters[:, 12] * (
            self.theta_scale[1] - self.theta_scale[0]
        )
        ci = self.ci_scale[0] + parameters[:, 13] * (
            self.ci_scale[1] - self.ci_scale[0]
        )
        cg = self.cg_scale[0] + parameters[:, 14] * (
            self.cg_scale[1] - self.cg_scale[0]
        )

        # initialize state values
        warmup_length = self.warmup_length
        if warmup_length > 0:
            # set no_grad for warmup periods
            with torch.no_grad():
                p_and_e_warmup = p_and_e[0:warmup_length, :, :]
                if self.param_test_way == MODEL_PARAM_TEST_WAY["time_varying"]:
                    parameters_ts_warmup = parameters_ts[0:warmup_length, :, :]
                else:
                    parameters_ts_warmup = parameters_ts
                cal_init_xaj4dpl = Xaj4DplWithNnModule(
                    kernel_size=self.kernel_size,
                    # warmup_length must be 0 here
                    warmup_length=0,
                    nn_module=self.evap_nn_module,
                    param_var_index=self.param_var_index,
                    source_book=self.source_book,
                    source_type=self.source_type,
                    et_output=self.et_output,
                    nn_hidden_size=self.nn_hidden_size,
                    nn_dropout=self.nn_dropout,
                    param_test_way=self.param_test_way,
                )
                if cal_init_xaj4dpl.warmup_length > 0:
                    raise RuntimeError("Please set init model's warmup length to 0!!!")
                _, _, *w0, s0, fr0, qi0, qg0 = cal_init_xaj4dpl(
                    p_and_e_warmup, parameters_ts_warmup, return_state=True
                )
        else:
            # use detach func to make wu0 no_grad as it is an initial value
            if self.et_output == 1:
                # () and , must be added, otherwise, w0 will be a tensor, not a tuple
                w0 = (0.5 * (um.detach() + lm.detach() + dm.detach()),)
            else:
                raise ValueError("et_output should be 1 or 3")
            s0 = 0.5 * (sm.detach())
            fr0 = torch.full(ci.size(), 0.1).to(xaj_device)
            qi0 = torch.full(ci.size(), 0.1).to(xaj_device)
            qg0 = torch.full(cg.size(), 0.1).to(xaj_device)

        inputs = p_and_e[warmup_length:, :, :]
        runoff_ims_ = torch.full(inputs.shape[:2], 0.0).to(xaj_device)
        rss_ = torch.full(inputs.shape[:2], 0.0).to(xaj_device)
        ris_ = torch.full(inputs.shape[:2], 0.0).to(xaj_device)
        rgs_ = torch.full(inputs.shape[:2], 0.0).to(xaj_device)
        es_ = torch.full(inputs.shape[:2], 0.0).to(xaj_device)
        for i in range(inputs.shape[0]):
            if 0 in self.param_var_index or self.param_var_index is None:
                k = ks[i]
            else:
                k = ks
            if 1 in self.param_var_index or self.param_var_index is None:
                b = bs[i]
            else:
                b = bs
            if 6 in self.param_var_index or self.param_var_index is None:
                c = cs[i]
            else:
                c = cs
            if i == 0:
                (r, rim, e, pe), w = self.xaj_generation_with_new_module(
                    inputs[i, :, :], k, b, im, um, lm, dm, c, *w0
                )
                if self.source_type == "sources":
                    (rs, ri, rg), (s, fr) = xaj_sources(
                        pe, r, sm, ex, ki, kg, s0, fr0, book=self.source_book
                    )
                elif self.source_type == "sources5mm":
                    (rs, ri, rg), (s, fr) = xaj_sources5mm(
                        pe, r, sm, ex, ki, kg, s0, fr0, book=self.source_book
                    )
                else:
                    raise NotImplementedError("No such divide-sources method")
            else:
                (r, rim, e, pe), w = self.xaj_generation_with_new_module(
                    inputs[i, :, :], k, b, im, um, lm, dm, c, *w
                )
                if self.source_type == "sources":
                    (rs, ri, rg), (s, fr) = xaj_sources(
                        pe, r, sm, ex, ki, kg, s, fr, book=self.source_book
                    )
                elif self.source_type == "sources5mm":
                    (rs, ri, rg), (s, fr) = xaj_sources5mm(
                        pe, r, sm, ex, ki, kg, s, fr, book=self.source_book
                    )
                else:
                    raise NotImplementedError("No such divide-sources method")
            # impevious part is pe * im
            runoff_ims_[i, :] = rim
            # so for non-imprvious part, the result should be corrected
            rss_[i, :] = rs * (1 - im)
            ris_[i, :] = ri * (1 - im)
            rgs_[i, :] = rg * (1 - im)
            es_[i, :] = e
        # seq, batch, feature
        runoff_im = torch.unsqueeze(runoff_ims_, dim=2)
        rss = torch.unsqueeze(rss_, dim=2)
        es = torch.unsqueeze(es_, dim=2)

        conv_uh = KernelConv(a, theta, self.kernel_size)
        qs_ = conv_uh(runoff_im + rss)

        qs = torch.full(inputs.shape[:2], 0.0).to(xaj_device)
        for i in range(inputs.shape[0]):
            if i == 0:
                qi = linear_reservoir(ris_[i], ci, qi0)
                qg = linear_reservoir(rgs_[i], cg, qg0)
            else:
                qi = linear_reservoir(ris_[i], ci, qi)
                qg = linear_reservoir(rgs_[i], cg, qg)
            qs[i, :] = qs_[i, :, 0] + qi + qg
        # seq, batch, feature
        q_sim = torch.unsqueeze(qs, dim=2)
        if return_state:
            return q_sim, es, *w, s, fr, qi, qg
        return q_sim, es


class DplLstmNnModuleXaj(nn.Module):
    def __init__(
        self,
        n_input_features,
        n_output_features,
        n_hidden_states,
        kernel_size,
        warmup_length,
        param_limit_func="clamp",
        param_test_way="final",
        param_var_index=None,
        source_book="HF",
        source_type="sources",
        nn_hidden_size=None,
        nn_dropout=0.2,
        et_output=3,
    ):
        """
        Differential Parameter learning model: LSTM -> Param -> XAJ

        The principle can be seen here: https://doi.org/10.1038/s41467-021-26107-z

        Parameters
        ----------
        n_input_features
            the number of input features of LSTM
        n_output_features
            the number of output features of LSTM, and it should be equal to the number of learning parameters in XAJ
        n_hidden_states
            the number of hidden features of LSTM
        kernel_size
            the time length of unit hydrograph
        warmup_length
            the length of warmup periods;
            hydrologic models need a warmup period to generate reasonable initial state values
        param_limit_func
            function used to limit the range of params; now it is sigmoid or clamp function
        param_test_way
            how we use parameters from dl model when testing;
            now we have three ways:
            1. "final" -- use the final period's parameter for each period
            2. "mean_time" -- Mean values of all periods' parameters is used
            3. "mean_basin" -- Mean values of all basins' final periods' parameters is used
            but remember these ways are only for non-variable parameters
        param_var_index
            variable parameters' indices in all parameters
        """
        if param_var_index is None:
            param_var_index = [0, 1, 6]
        if nn_hidden_size is None:
            nn_hidden_size = [16, 8]
        super(DplLstmNnModuleXaj, self).__init__()
        self.dl_model = SimpleLSTM(n_input_features, n_output_features, n_hidden_states)
        self.pb_model = Xaj4DplWithNnModule(
            kernel_size,
            warmup_length,
            source_book=source_book,
            source_type=source_type,
            nn_hidden_size=nn_hidden_size,
            nn_dropout=nn_dropout,
            et_output=et_output,
            param_var_index=param_var_index,
            param_test_way=param_test_way,
        )
        self.param_func = param_limit_func
        self.param_test_way = param_test_way
        self.param_var_index = param_var_index

    def forward(self, x, z):
        """
        Differential parameter learning

        z (normalized input) -> lstm -> param -> + x (not normalized) -> xaj -> q
        Parameters will be denormalized in xaj model

        Parameters
        ----------
        x
            not normalized data used for physical model; a sequence-first 3-dim tensor. [sequence, batch, feature]
        z
            normalized data used for DL model; a sequence-first 3-dim tensor. [sequence, batch, feature]

        Returns
        -------
        torch.Tensor
            one time forward result
        """
        gen = self.dl_model(z)
        if torch.isnan(gen).any():
            raise ValueError("Error: NaN values detected. Check your data firstly!!!")
        # we set all params' values in [0, 1] and will scale them when forwarding
        if self.param_func == "sigmoid":
            params = F.sigmoid(gen)
        elif self.param_func == "clamp":
            params = torch.clamp(gen, min=0.0, max=1.0)
        else:
            raise NotImplementedError(
                "We don't provide this way to limit parameters' range!! Please choose sigmoid or clamp"
            )
        # just get one-period values, here we use the final period's values,
        # when the MODEL_PARAM_TEST_WAY is not time_varing, we use the last period's values.
        if self.param_test_way != MODEL_PARAM_TEST_WAY["time_varying"]:
            params = params[-1, :, :]
        # Please put p in the first location and pet in the second
        q, e = self.pb_model(x[:, :, : self.pb_model.feature_size], params)
        return torch.cat([q, e], dim=-1)



"""
Author: Wenyu Ouyang
Date: 2024-05-01 10:34:15
LastEditTime: 2024-05-04 11:31:52
LastEditors: Wenyu Ouyang
Description: Unit tests for the DeepHydro class
FilePath: \torchhydro\tests\test_deep_hydro.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import pytest
import os
import torch
from torch.utils.data import Dataset
from torchhydro.trainers.deep_hydro import DeepHydro
from torchhydro.datasets.data_dict import datasets_dict
from torchhydro.trainers.train_logger import TrainLogger
import torch
import tempfile


# Mock dataset class using random data
class MockDataset(Dataset):
    def __init__(self, data_cfgs, is_tra_val_te):
        super(MockDataset, self).__init__()
        self.data_cfgs = data_cfgs  # Store the passed configuration for later use
        # Simulate other configuration and setup steps

    @property
    def ngrid(self):
        return 10

    @property
    def nt(self):
        return 200

    def __len__(self):
        return self.ngrid * (self.nt - self.data_cfgs["forecast_history"] + 1)

    def __getitem__(self, idx):
        # Use the stored configurations to generate mock data
        # rho = self.data_cfgs["forecast_history"]
        # x = torch.randn(rho, self.data_cfgs["input_features"])
        # y = torch.randn(rho, self.data_cfgs["output_features"])
        mode = self.data_cfgs["model_mode"]
        if mode == "single":
            src1 = torch.rand(
                self.data_cfgs["forecast_history"],
                self.data_cfgs["input_features"] - 1,
            )
            src2 = torch.rand(
                self.data_cfgs["forecast_history"],
                self.data_cfgs["cnn_size"],
            )
        else:
            src1 = torch.rand(
                self.data_cfgs["forecast_history"],
                self.data_cfgs["input_features"],
            )
            src2 = torch.rand(
                self.data_cfgs["forecast_history"],
                self.data_cfgs["input_size_encoder2"]
                )
        src3 = torch.rand(1, self.data_cfgs["output_features"]) # start_token
        out = torch.rand(self.data_cfgs["forecast_length"], self.data_cfgs["output_features"]) 
        return [src1, src2, src3], out


@pytest.fixture()
def dummy_data_cfgs():
    test_path = "results/test_seq2seq_single2/"
    if not os.path.exists(test_path):
        os.makedirs(test_path)
    return {
        "dataset": "MockDataset",
        "input_features": 4,
        "output_features": 1,
        # "t_range_valid": ["2010-01-01", "2010-12-31"],
        "t_range_valid": None,
        "test_path": test_path,
        "sampler": "KuaiSampler",
        # "sampler": "HydroSampler",
        "batch_size": 5,
        "forecast_history": 5,
        "forecast_length": 2,
        "warmup_length": 10,
        "cnn_size" : 120,
        "input_size_encoder2": 1,
        "model_mode": "single",
    }


def test_using_mock_dataset(dummy_data_cfgs):
    datasets_dict["MockDataset"] = MockDataset
    is_tra_val_te = True
    dataset_name = "MockDataset"

    dataset = datasets_dict[dataset_name](dummy_data_cfgs, is_tra_val_te)

    assert len(dataset) == 330
    sample_x, sample_y= dataset[0]
    print(sample_x[0].shape)
    print(sample_x[1].shape)
    print(sample_x[2].shape)
    print(sample_y.shape)
    # assert sample_x.shape == (dummy_data_cfgs['forecast_history'], dummy_data_cfgs["input_features"])
    # assert sample_y.shape == (dummy_data_cfgs['forecast_history'], dummy_data_cfgs["output_features"])

@pytest.fixture()
def dummy_train_cfgs(dummy_data_cfgs):
    return {
        "training_cfgs": {
            "early_stopping": False,
            "patience": 4,
            "epochs": 2,
            "start_epoch": 1,
            "which_first_tensor": "batch",
            "device": -1,  # Assuming CPU device
            "train_mode": True,
            "criterion": "RMSE",
            "optimizer": "Adam",
            "optim_params": {},
            "lr_scheduler": {"lr": 0.001},
            "batch_size": 5,
            "save_epoch": 1,
        },
        "data_cfgs": dummy_data_cfgs,
        "model_cfgs": {
            "model_type": "Normal",
            # "model_name": "CpuLSTM",
            "model_name": "Seq2Seq",
            "weight_path": None,
            "model_hyperparam": {
                # "n_input_features": 10,
                # "n_output_features": 1,
                # "n_hidden_states": 64,
                "input_size": 4,
                "output_size": 1,
                "hidden_size": 256,
                "forecast_length": 2,
                "cnn_size": 120,
                "model_mode": "single",
            },
        },
        "evaluation_cfgs": {
            "model_loader": {"load_way": "specified", "test_epoch": 2},
            "metrics": ["NSE", "RMSE", "KGE"],
            "fill_nan": "no",
        },
    }

@pytest.fixture()
def deep_hydro(dummy_train_cfgs):
    datasets_dict["MockDataset"] = MockDataset
    # Create an instance of DeepHydro with dummy configurations
    return DeepHydro(dummy_train_cfgs)


def test_model_train(deep_hydro):
    # Call the model_train method and check if it runs without errors
    deep_hydro.model_train()

    # Add assertions to check the expected behavior of the method
    assert deep_hydro.model.state_dict() is not None

def test_plot_model_structure(deep_hydro, dummy_train_cfgs):
    opt = torch.optim.SGD(deep_hydro.model.parameters(), lr=0.01)
    model_filepath = dummy_train_cfgs["data_cfgs"]["test_path"]
    train_logger = TrainLogger(model_filepath, dummy_train_cfgs, opt)
    train_logger.plot_model_structure(deep_hydro.model)


"""
Author: MHPI group, Wenyu Ouyang
Date: 2021-12-31 11:08:29
LastEditTime: 2024-05-27 16:01:38
LastEditors: Wenyu Ouyang
Description: LSTM with dropout implemented by Kuai Fang and more LSTMs using it
FilePath: \torchhydro\torchhydro\models\cudnnlstm.py
Copyright (c) 2021-2022 MHPI group, Wenyu Ouyang. All rights reserved.
"""

import math
from typing import Union

import torch
import torch.nn as nn
from torch.nn import Parameter
import torch.nn.functional as F

from torchhydro.models.ann import SimpleAnn
from torchhydro.models.dropout import DropMask, create_mask


class LstmCellTied(nn.Module):
    """
    LSTM with dropout implemented by Kuai Fang: https://github.com/mhpi/hydroDL/blob/release/hydroDL/model/rnn.py

    the name of "Tied" comes from this paper:
    http://papers.nips.cc/paper/6241-a-theoretically-grounded-application-of-dropout-in-recurrent-neural-networks.pdf
    which means the weights of all gates will be tied together to be used (eq. 6 in this paper).
    this code is mainly used as a CPU version of CudnnLstm
    """

    def __init__(
        self,
        *,
        input_size,
        hidden_size,
        mode="train",
        dr=0.5,
        dr_method="drX+drW+drC",
        gpu=1
    ):
        super(LstmCellTied, self).__init__()

        self.inputSize = input_size
        self.hiddenSize = hidden_size
        self.dr = dr

        self.w_ih = Parameter(torch.Tensor(hidden_size * 4, input_size))
        self.w_hh = Parameter(torch.Tensor(hidden_size * 4, hidden_size))
        self.b_ih = Parameter(torch.Tensor(hidden_size * 4))
        self.b_hh = Parameter(torch.Tensor(hidden_size * 4))

        self.drMethod = dr_method.split("+")
        self.gpu = gpu
        self.mode = mode
        if mode == "train":
            self.train(mode=True)
        elif mode in ["test", "drMC"]:
            self.train(mode=False)
        if gpu >= 0:
            self = self.cuda()
            self.is_cuda = True
        else:
            self.is_cuda = False
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hiddenSize)
        for weight in self.parameters():
            weight.data.uniform_(-stdv, stdv)

    def reset_mask(self, x, h, c):
        self.mask_x = create_mask(x, self.dr)
        self.mask_h = create_mask(h, self.dr)
        self.mask_c = create_mask(c, self.dr)
        self.mask_w_ih = create_mask(self.w_ih, self.dr)
        self.mask_w_hh = create_mask(self.w_hh, self.dr)

    def forward(self, x, hidden, *, do_reset_mask=True, do_drop_mc=False):
        do_drop = self.dr > 0 and (do_drop_mc is True or self.training is True)
        batch_size = x.size(0)
        h0, c0 = hidden
        if h0 is None:
            h0 = x.new_zeros(batch_size, self.hiddenSize, requires_grad=False)
        if c0 is None:
            c0 = x.new_zeros(batch_size, self.hiddenSize, requires_grad=False)

        if self.dr > 0 and self.training is True and do_reset_mask is True:
            self.reset_mask(x, h0, c0)

        if do_drop and "drH" in self.drMethod:
            h0 = DropMask.apply(h0, self.mask_h, True)

        if do_drop and "drX" in self.drMethod:
            x = DropMask.apply(x, self.mask_x, True)

        if do_drop and "drW" in self.drMethod:
            w_ih = DropMask.apply(self.w_ih, self.mask_w_ih, True)
            w_hh = DropMask.apply(self.w_hh, self.mask_w_hh, True)
        else:
            # self.w are parameters, while w are not
            w_ih = self.w_ih
            w_hh = self.w_hh

        gates = F.linear(x, w_ih, self.b_ih) + F.linear(h0, w_hh, self.b_hh)
        gate_i, gate_f, gate_c, gate_o = gates.chunk(4, 1)

        gate_i = torch.sigmoid(gate_i)
        gate_f = torch.sigmoid(gate_f)
        gate_c = torch.tanh(gate_c)
        gate_o = torch.sigmoid(gate_o)

        if self.training is True and "drC" in self.drMethod:
            gate_c = gate_c.mul(self.mask_c)

        c1 = (gate_f * c0) + (gate_i * gate_c)
        h1 = gate_o * torch.tanh(c1)

        return h1, c1


class CpuLstmModel(nn.Module):
    """Cpu version of CudnnLstmModel"""

    def __init__(self, *, n_input_features, n_output_features, n_hidden_states, dr=0.5):
        super(CpuLstmModel, self).__init__()
        self.nx = n_input_features
        self.ny = n_output_features
        self.hiddenSize = n_hidden_states
        self.ct = 0
        self.nLayer = 1
        self.linearIn = torch.nn.Linear(n_input_features, n_hidden_states)
        self.lstm = LstmCellTied(
            input_size=n_hidden_states,
            hidden_size=n_hidden_states,
            dr=dr,
            dr_method="drW",
            gpu=-1,
        )
        self.linearOut = torch.nn.Linear(n_hidden_states, n_output_features)
        self.gpu = -1

    def forward(self, x, do_drop_mc=False):
        # x0 = F.relu(self.linearIn(x))
        # outLSTM, (hn, cn) = self.lstm(x0, do_drop_mc=do_drop_mc)
        # out = self.linearOut(outLSTM)
        # return out
        nt, ngrid, nx = x.shape
        yt = torch.zeros(ngrid, 1)
        out = torch.zeros(nt, ngrid, self.ny)
        ht = None
        ct = None
        reset_mask = True
        for t in range(nt):
            xt = x[t, :, :]
            xt = torch.where(torch.isnan(xt), torch.full_like(xt, 0), xt)
            x0 = F.relu(self.linearIn(xt))
            ht, ct = self.lstm(x0, hidden=(ht, ct), do_reset_mask=reset_mask)
            yt = self.linearOut(ht)
            reset_mask = False
            out[t, :, :] = yt
        return out


class CudnnLstm(nn.Module):
    """
    LSTM with dropout implemented by Kuai Fang: https://github.com/mhpi/hydroDL/blob/release/hydroDL/model/rnn.py

    Only run in GPU; the CPU version is LstmCellTied in this file
    """

    def __init__(self, *, input_size, hidden_size, dr=0.5):
        """

        Parameters
        ----------
        input_size
            number of neurons in input layer
        hidden_size
            number of neurons in hidden layer
        dr
            dropout rate
        """
        super(CudnnLstm, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.dr = dr
        self.w_ih = Parameter(torch.Tensor(hidden_size * 4, input_size))
        self.w_hh = Parameter(torch.Tensor(hidden_size * 4, hidden_size))
        self.b_ih = Parameter(torch.Tensor(hidden_size * 4))
        self.b_hh = Parameter(torch.Tensor(hidden_size * 4))
        self._all_weights = [["w_ih", "w_hh", "b_ih", "b_hh"]]
        # self.cuda()

        self.reset_mask()
        self.reset_parameters()

    def _apply(self, fn):
        return super()._apply(fn)

    def __setstate__(self, d):  # this func will be called when loading the model
        super().__setstate__(d)
        self.__dict__.setdefault("_data_ptrs", [])
        if "all_weights" in d:
            self._all_weights = d["all_weights"]
        if isinstance(self._all_weights[0][0], str):
            return
        self._all_weights = [["w_ih", "w_hh", "b_ih", "b_hh"]]

    def reset_mask(self):
        self.mask_w_ih = create_mask(self.w_ih, self.dr)
        self.mask_w_hh = create_mask(self.w_hh, self.dr)

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.hidden_size)
        for weight in self.parameters():
            weight.data.uniform_(-stdv, stdv)

    def forward(self, input, hx=None, cx=None, do_drop_mc=False, dropout_false=False):
        # dropout_false: it will ensure do_drop is false, unless do_drop_mc is true
        if dropout_false and (not do_drop_mc):
            do_drop = False
        elif self.dr > 0 and (do_drop_mc is True or self.training is True):
            do_drop = True
        else:
            do_drop = False

        batch_size = input.size(1)

        if hx is None:
            hx = input.new_zeros(1, batch_size, self.hidden_size, requires_grad=False)
        if cx is None:
            cx = input.new_zeros(1, batch_size, self.hidden_size, requires_grad=False)

        # handle = torch.backends.cudnn.get_handle()
        if do_drop is True:
            # cuDNN backend - disabled flat weight
            freeze_mask = False
            if not freeze_mask:
                self.reset_mask()
            weight = [
                DropMask.apply(self.w_ih, self.mask_w_ih, True),
                DropMask.apply(self.w_hh, self.mask_w_hh, True),
                self.b_ih,
                self.b_hh,
            ]
        else:
            weight = [self.w_ih, self.w_hh, self.b_ih, self.b_hh]
        if torch.__version__ < "1.8":
            output, hy, cy, reserve, new_weight_buf = torch._cudnn_rnn(
                input,
                weight,
                4,
                None,
                hx,
                cx,
                2,  # 2 means LSTM
                self.hidden_size,
                1,
                False,
                0,
                self.training,
                False,
                (),
                None,
            )
        else:
            output, hy, cy, reserve, new_weight_buf = torch._cudnn_rnn(
                input,
                weight,
                4,
                None,
                hx,
                cx,
                2,  # 2 means LSTM
                self.hidden_size,
                0,
                1,
                False,
                0,
                self.training,
                False,
                (),
                None,
            )
        return output, (hy, cy)

    @property
    def all_weights(self):
        return [
            [getattr(self, weight) for weight in weights]
            for weights in self._all_weights
        ]


class CudnnLstmModel(nn.Module):
    def __init__(self, n_input_features, n_output_features, n_hidden_states, dr=0.5):
        """
        An LSTM model writen by Kuai Fang from this paper: https://doi.org/10.1002/2017GL075619

        only gpu version

        Parameters
        ----------
        n_input_features
            the number of input features
        n_output_features
            the number of output features
        n_hidden_states
            the number of hidden features
        dr
            dropout rate and its default is 0.5
        """
        super(CudnnLstmModel, self).__init__()
        self.nx = n_input_features
        self.ny = n_output_features
        self.hidden_size = n_hidden_states
        self.ct = 0
        self.nLayer = 1
        self.linearIn = torch.nn.Linear(self.nx, self.hidden_size)
        self.lstm = CudnnLstm(
            input_size=self.hidden_size, hidden_size=self.hidden_size, dr=dr
        )
        self.linearOut = torch.nn.Linear(self.hidden_size, self.ny)

    def forward(self, x, do_drop_mc=False, dropout_false=False, return_h_c=False):
        x0 = F.relu(self.linearIn(x))
        out_lstm, (hn, cn) = self.lstm(
            x0, do_drop_mc=do_drop_mc, dropout_false=dropout_false
        )
        out = self.linearOut(out_lstm)
        return (out, (hn, cn)) if return_h_c else out


class LinearCudnnLstmModel(CudnnLstmModel):
    """This model is nonlinear layer + CudnnLSTM/CudnnLstm-MultiOutput-Model.
    kai_tl: model from this paper by Ma et al. -- https://doi.org/10.1029/2020WR028600
    """

    def __init__(self, linear_size, **kwargs):
        """

        Parameters
        ----------
        linear_size
            the number of input features for the first input linear layer
        """
        super(LinearCudnnLstmModel, self).__init__(**kwargs)
        self.former_linear = torch.nn.Linear(linear_size, kwargs["n_input_features"])

    def forward(self, x, do_drop_mc=False, dropout_false=False):
        x0 = F.relu(self.former_linear(x))
        return super(LinearCudnnLstmModel, self).forward(
            x0, do_drop_mc=do_drop_mc, dropout_false=dropout_false
        )


def cal_conv_size(lin, kernel, stride, padding=0, dilation=1):
    lout = (lin + 2 * padding - dilation * (kernel - 1) - 1) / stride + 1
    return int(lout)


def cal_pool_size(lin, kernel, stride=None, padding=0, dilation=1):
    if stride is None:
        stride = kernel
    lout = (lin + 2 * padding - dilation * (kernel - 1) - 1) / stride + 1
    return int(lout)


class CNN1dKernel(torch.nn.Module):
    def __init__(self, *, ninchannel=1, nkernel=3, kernelSize=3, stride=1, padding=0):
        super(CNN1dKernel, self).__init__()
        self.cnn1d = torch.nn.Conv1d(
            in_channels=ninchannel,
            out_channels=nkernel,
            kernel_size=kernelSize,
            padding=padding,
            stride=stride,
        )
        self.name = "CNN1dkernel"
        self.is_legacy = True

    def forward(self, x):
        return F.relu(self.cnn1d(x))


class CNN1dLCmodel(nn.Module):
    # Directly add the CNN extracted features into LSTM inputSize
    def __init__(
        self,
        nx,
        ny,
        nobs,
        hidden_size,
        n_kernel: Union[list, tuple] = (10, 5),
        kernel_size: Union[list, tuple] = (3, 3),
        stride: Union[list, tuple] = (2, 1),
        dr=0.5,
        pool_opt=None,
        cnn_dr=0.5,
        cat_first=True,
    ):
        """cat_first means: we will concatenate the CNN output with the x, then input them to the CudnnLstm model;
        if not cat_first, it is relu_first, meaning we will relu the CNN output firstly, then concatenate it with x
        """
        # two convolutional layer
        super(CNN1dLCmodel, self).__init__()
        # N_cnn_out代表输出的特征数量
        # nx代表历史的输入
        # ny代表最后线性层输出的维度，如果只预报流量，则为1
        # nobs代表要输入到CNN的维度
        # hidden_size是线性层的隐藏层的节点数
        self.nx = nx
        self.ny = ny
        self.obs = nobs
        self.hiddenSize = hidden_size
        n_layer = len(n_kernel)
        self.features = nn.Sequential()
        n_in_chan = 1
        lout = nobs
        for ii in range(n_layer):
            conv_layer = CNN1dKernel(
                ninchannel=n_in_chan,
                nkernel=n_kernel[ii],
                kernelSize=kernel_size[ii],
                stride=stride[ii],
            )
            self.features.add_module("CnnLayer%d" % (ii + 1), conv_layer)
            if cnn_dr != 0.0:
                self.features.add_module("dropout%d" % (ii + 1), nn.Dropout(p=cnn_dr))
            n_in_chan = n_kernel[ii]
            lout = cal_conv_size(lin=lout, kernel=kernel_size[ii], stride=stride[ii])
            self.features.add_module("Relu%d" % (ii + 1), nn.ReLU())
            if pool_opt is not None:
                self.features.add_module(
                    "Pooling%d" % (ii + 1), nn.MaxPool1d(pool_opt[ii])
                )
                lout = cal_pool_size(lin=lout, kernel=pool_opt[ii])
        self.N_cnn_out = int(
            lout * n_kernel[-1]
        )  # total CNN feature number after convolution
        self.cat_first = cat_first
        # 要不要先拼接？
        # 先拼接，则代表线性层中，输入的维度是未来的降水等输入输出的CNN特征维度，和历史观测的等时间序列的特征数量，通过线性层合并成一个，然后再把这些特征输出到一个线性层中
        # 如果不拼接，那么历史观测数据先进入一个线性层
        if cat_first:
            nf = self.N_cnn_out + nx
            self.linearIn = torch.nn.Linear(nf, hidden_size)
            # CudnnLstm除了最基础的部分以外，主要是有个h和c两个门为空的纠错，这个在论文里讲述的是因为可能输入缺失，但是又不想用插值处理
            # 不想用插值处理是因为认为会暴露未来信息
            # 采用了置零操作，原文的表述是这种缺失点较少，在模型的不断更新参数后，这种置零的影响对于模型的输出影响很小
            self.lstm = CudnnLstm(
                input_size=hidden_size, hidden_size=hidden_size, dr=dr
            )
        else:
            nf = self.N_cnn_out + hidden_size
            self.linearIn = torch.nn.Linear(nx, hidden_size)
            self.lstm = CudnnLstm(input_size=nf, hidden_size=hidden_size, dr=dr)
        self.linearOut = torch.nn.Linear(hidden_size, ny)
        self.gpu = 1

    def forward(self, x, z, do_drop_mc=False):
        # z = n_grid*nVar add a channel dimension
        # z = z.t()
        n_grid, nobs, _ = z.shape
        z = z.reshape(n_grid * nobs, 1)
        n_t, bs, n_var = x.shape
        # add a channel dimension
        z = torch.unsqueeze(z, dim=1)
        z0 = self.features(z)
        # z0 = (n_grid) * n_kernel * sizeafterconv
        z0 = z0.view(n_grid, self.N_cnn_out).repeat(n_t, 1, 1)
        if self.cat_first:
            x = torch.cat((x, z0), dim=2)
            x0 = F.relu(self.linearIn(x))
        else:
            x = F.relu(self.linearIn(x))
            x0 = torch.cat((x, z0), dim=2)
        out_lstm, (hn, cn) = self.lstm(x0, do_drop_mc=do_drop_mc)
        return self.linearOut(out_lstm)


class CudnnLstmModelLstmKernel(nn.Module):
    """use a trained/un-trained CudnnLstm as a kernel generator before another CudnnLstm."""

    def __init__(
        self,
        nx,
        ny,
        hidden_size,
        nk=None,
        hidden_size_later=None,
        cut=False,
        dr=0.5,
        delta_s=False,
    ):
        """delta_s means we will use the difference of the first lstm's output and the second's as the final output"""
        super(CudnnLstmModelLstmKernel, self).__init__()
        # These three layers are same with CudnnLstmModel to be used for transfer learning or just vanilla-use
        self.linearIn = torch.nn.Linear(nx, hidden_size)
        self.lstm = CudnnLstm(input_size=hidden_size, hidden_size=hidden_size, dr=dr)
        self.linearOut = torch.nn.Linear(hidden_size, ny)
        # if cut is True, we will only select the final index in nk, and repeat it, then concatenate with x
        self.cut = cut
        # the second lstm has more input than the previous
        if nk is None:
            nk = ny
        if hidden_size_later is None:
            hidden_size_later = hidden_size
        self.linear_in_later = torch.nn.Linear(nx + nk, hidden_size_later)
        self.lstm_later = CudnnLstm(
            input_size=hidden_size_later, hidden_size=hidden_size_later, dr=dr
        )
        self.linear_out_later = torch.nn.Linear(hidden_size_later, ny)

        self.delta_s = delta_s
        # when delta_s is true, cut cannot be true, because they have to have same number params
        assert not (cut and delta_s)

    def forward(self, x, do_drop_mc=False, dropout_false=False):
        x0 = F.relu(self.linearIn(x))
        out_lstm1, (hn1, cn1) = self.lstm(
            x0, do_drop_mc=do_drop_mc, dropout_false=dropout_false
        )
        gen = self.linearOut(out_lstm1)
        if self.cut:
            gen = gen[-1, :, :].repeat(x.shape[0], 1, 1)
        x1 = torch.cat((x, gen), dim=len(gen.shape) - 1)
        x2 = F.relu(self.linear_in_later(x1))
        out_lstm2, (hn2, cn2) = self.lstm_later(
            x2, do_drop_mc=do_drop_mc, dropout_false=dropout_false
        )
        out = self.linear_out_later(out_lstm2)
        return gen - out if self.delta_s else (out, gen)


class CudnnLstmModelMultiOutput(nn.Module):
    def __init__(
        self,
        n_input_features,
        n_output_features,
        n_hidden_states,
        layer_hidden_size=(128, 64),
        dr=0.5,
        dr_hidden=0.0,
    ):
        """
        Multiple output CudnnLSTM.

        It has multiple output layers, each for one output, so that we can easily freeze any output layer.

        Parameters
        ----------
        n_input_features
            the size of input features
        n_output_features
            the size of output features; in this model, we set different nonlinear layer for each output
        n_hidden_states
            the size of LSTM's hidden features
        layer_hidden_size
            hidden_size for multi-layers
        dr
            dropout rate
        dr_hidden
            dropout rates of hidden layers
        """
        super(CudnnLstmModelMultiOutput, self).__init__()
        self.ct = 0
        multi_layers = torch.nn.ModuleList()
        for i in range(n_output_features):
            multi_layers.add_module(
                "layer%d" % (i + 1),
                SimpleAnn(n_hidden_states, 1, layer_hidden_size, dr=dr_hidden),
            )
        self.multi_layers = multi_layers
        self.linearIn = torch.nn.Linear(n_input_features, n_hidden_states)
        self.lstm = CudnnLstm(
            input_size=n_hidden_states, hidden_size=n_hidden_states, dr=dr
        )

    def forward(self, x, do_drop_mc=False, dropout_false=False, return_h_c=False):
        x0 = F.relu(self.linearIn(x))
        out_lstm, (hn, cn) = self.lstm(
            x0, do_drop_mc=do_drop_mc, dropout_false=dropout_false
        )
        outs = [mod(out_lstm) for mod in self.multi_layers]
        final = torch.cat(outs, dim=-1)
        return (final, (hn, cn)) if return_h_c else final



"""
Author: Wenyu Ouyang
Date: 2021-12-31 11:08:29
LastEditTime: 2024-09-15 10:04:38
LastEditors: Wenyu Ouyang
Description: Config for hydroDL
FilePath: \torchhydro\torchhydro\configs\config.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

import argparse
import fnmatch
import json
import os

import numpy as np
import pandas as pd
from hydroutils import hydro_file

DAYMET_NAME = "daymet"
SSM_SMAP_NAME = "ssm"
ET_MODIS_NAME = "ET"
Q_CAMELS_US_NAME = "streamflow"
Q_CAMELS_CC_NAME = "Q"
PRCP_DAYMET_NAME = "prcp"
PRCP_NLDAS_NAME = "total_precipitation"
PET_MODIS_NAME = "PET"
PET_NLDAS_NAME = "potential_evaporation"
NLDAS_NAME = "nldas"
ERA5LAND_NAME = "era5land"
ET_ERA5LAND_NAME = "total_evaporation"
PRCP_ERA5LAND_NAME = "total_precipitation"
PET_DAYMET_NAME = "PET"
PET_ERA5LAND_NAME = "potential_evaporation"
DATE_FORMATS = ["%Y-%m-%d-%H", "%Y-%m-%d"]  # 带小时的日期格式  # 不带小时的日期格式


def default_config_file():
    """
    Default config file for all models/data/training parameters in this repo

    Returns
    -------
    dict
        configurations
    """

    return {
        "model_cfgs": {
            # model_type including normal deep learning (Normal), federated learning (FedLearn), transfer learing (TransLearn), multi-task learning (MTL), etc.
            "model_type": "Normal",
            # supported models can be seen in hydroDL/model_dict_function.py
            "model_name": "LSTM",
            # the details of model parameters for the "model_name" model
            "model_hyperparam": {
                # <- warmup -><- forecast_history -><- forecast_length ->
                "forecast_history": 30,
                "forecast_length": 30,
                # the size of input (feature number)
                "input_size": 24,
                # the length of output time-sequence (feature number)
                "output_size": 1,
                "hidden_size": 20,
                "num_layers": 1,
                "bias": True,
                "batch_size": 100,
                "dropout": 0.2,
            },
            "weight_path": None,
            "continue_train": True,
            # federated learning parameters
            "fl_hyperparam": {
                # sampling for federated learning
                "fl_sample": "basin",
                # number of users for federated learning
                # TODO: we don't use this parameter now, but we may use it in the future
                "fl_num_users": 10,
                # the number of local epochs
                "fl_local_ep": 5,
                # local batch size
                "fl_local_bs": 6,
                # the fraction of clients
                "fl_frac": 0.1,
            },
            "tl_hyperparam": {
                # part of transfer learning in a model: a list of layers' names, such as ["lstm"]
                "tl_part": None,
            },
        },
        "data_cfgs": {
            "source_cfgs": {
                # the name of data source, such as CAMELS
                "source_names": ["CAMELS"],
                "source_paths": ["../../example/camels_us"],
            },
            "validation_path": None,
            "test_path": None,
            "batch_size": 100,
            # we generally have three times: [warmup, forecast_history, forecast_length]
            # For physics-based models, we need warmup; default is 0 as DL models generally don't need it
            "warmup_length": 0,
            # the length of the history data for forecasting
            "forecast_history": 30,
            # the length of the forecast data
            "forecast_length": 1,
            # the min time step of the input data
            "min_time_unit": "D",
            # the min time interval of the input data
            "min_time_interval": 1,
            # modeled objects
            "object_ids": "ALL",
            # modeling time range
            "t_range_train": ["1992-01-01", "1993-01-01"],
            "t_range_valid": None,
            "t_range_test": ["1993-01-01", "1994-01-01"],
            # the output
            "target_cols": [Q_CAMELS_US_NAME],
            "target_rm_nan": True,
            # only for cases in which target data will be used as input:
            # data assimilation -- use streamflow from period 0 to t-1 (TODO: not included now)
            # for physics-based model -- use streamflow to calibrate models
            "target_as_input": False,
            # the time series input
            # TODO: now we only support one forcing type
            "relevant_types": [DAYMET_NAME],
            "relevant_cols": [
                "dayl",
                PRCP_DAYMET_NAME,
                "srad",
                "swe",
                "tmax",
                "tmin",
                "vp",
            ],
            "relevant_rm_nan": True,
            # the attribute input
            "constant_cols": [
                "elev_mean",
                "slope_mean",
                "area_gages2",
                "frac_forest",
                "lai_max",
                "lai_diff",
                "dom_land_cover_frac",
                "dom_land_cover",
                "root_depth_50",
                "soil_depth_statsgo",
                "soil_porosity",
                "soil_conductivity",
                "max_water_content",
                "geol_1st_class",
                "geol_2nd_class",
                "geol_porostiy",
                "geol_permeability",
            ],
            # specify the data source of each variable
            "var_to_source_map": None,
            # {
            #     "temperature": "nldas4camels",
            #     "specific_humidity": "nldas4camels",
            #     "usgsFlow": "camels_us",
            #     "ET": "modiset4camels",
            # },
            "constant_rm_nan": True,
            # if constant_only, we will only use constant data as DL models' input: this is only for dpl models now
            "constant_only": False,
            # more other cols, use dict to express!
            "other_cols": None,
            # only numerical scaler: for categorical vars, they are transformed to numerical vars when reading them
            "scaler": "StandardScaler",
            # Some parameters for the chosen scaler function, default is DapengScaler's
            "scaler_params": {
                "prcp_norm_cols": [
                    Q_CAMELS_US_NAME,
                    "streamflow",
                    Q_CAMELS_CC_NAME,
                    "qobs",
                    "qobs_mm_per_hour",
                ],
                "gamma_norm_cols": [
                    PRCP_DAYMET_NAME,
                    "pr",
                    # PRCP_ERA5LAND_NAME is same as PRCP_NLDAS_NAME
                    PRCP_NLDAS_NAME,
                    "pre",
                    # pet may be negative, but we set negative as 0 because of gamma_norm_cols
                    # https://earthscience.stackexchange.com/questions/12031/does-negative-reference-evapotranspiration-make-sense-using-fao-penman-monteith
                    "pet",
                    # PET_ERA5LAND_NAME is same as PET_NLDAS_NAME
                    PET_NLDAS_NAME,
                    ET_MODIS_NAME,
                    "LE",
                    PET_MODIS_NAME,
                    "PLE",
                    "GPP",
                    "Ec",
                    "Es",
                    "Ei",
                    "ET_water",
                    "ET_sum",
                    SSM_SMAP_NAME,
                    "susm",
                    "smp",
                    "ssma",
                    "susma",
                ],
                "pbm_norm": False,
            },
            "stat_dict_file": None,
            # dataset for pytorch dataset
            "dataset": "StreamflowDataset",
            # sampler for pytorch dataloader, here we mainly use it for Kuai Fang's sampler in all his DL papers
            "sampler": None,
        },
        "training_cfgs": {
            "master_addr": "localhost",
            "port": "12335",
            # if train_mode is False, don't train and evaluate
            "train_mode": True,
            "criterion": "RMSE",
            "criterion_params": None,
            # "weight_decay": None, a regularization term in loss func
            "optimizer": "Adam",
            "optim_params": {},
            "lr_scheduler": {
                # 1st opt config, all epochs use this lr
                "lr": 0.001,
                # 2nd opt config, diff epoch uses diff lr, key is epoch,
                # start from 0, each value means the decay rate
                # "lr_scheduler": {0: 1, 1: 0.5, 2: 0.2},
                # 3rd opt config, lr as a initial value, and lr_factor as an exponential decay factor
                # "lr": 0.001, "lr_factor": 0.1,
                # 4th opt config, lr as a initial value,
                # lr_patience represent how many epochs without opt (we watch val_loss) could be tolerated
                # if lr_patience is satisfied, then lr will be decayed by lr_factor by a linear way
                # "lr": 0.001, "lr_factor": 0.1, "lr_patience": 1,
            },
            "early_stopping": False,
            "patience": 1,
            "epochs": 20,
            # save_epoch ==0 means only save once in the final epoch
            "save_epoch": 0,
            # save_iter ==0 means we don't save model during training in a epoch
            "save_iter": 0,
            # when we train a model for long time, some accidents may interrupt our training.
            # Then we need retrain the model with saved weights, and the start_epoch is not 1 yet.
            "start_epoch": 1,
            "batch_size": 100,
            "random_seed": 1234,
            "device": [0, 1, 2],
            "multi_targets": 1,
            "num_workers": 0,
            "which_first_tensor": "sequence",
            # for ensemble exp:
            # basically we set kfold/seeds/hyper_params for trianing such as batch_sizes
            "ensemble": False,
            "ensemble_items": {
                # kfold means a time cross validation,
                # concatenate train ,valid, and test data together,
                # then split them into k folds
                "kfold": None,
                "batch_sizes": None,
                # if seeds is not None,
                # we will use different seeds for different sub-exps
                "seeds": None,
                "patience": None,
                "early_stopping": None,
                "kfold_continuous": True,
            },
        },
        # For evaluation
        "evaluation_cfgs": {
            # there are some different loading way of trained model weights
            # 'epoch' means we load the weights of the specified epoch;
            # 'best' means we load the best weights during training especially for early stopping
            # 'latest' means we load the latest weights during training
            # 'pth' means we load the weights from the specified path
            "model_loader": {"load_way": "specified", "test_epoch": 20},
            # "model_loader": {"load_way": "best"},
            # "model_loader": {"load_way": "latest"},
            # "model_loader": {"load_way": "pth", "pth": "path/to/weights"},
            "metrics": ["NSE", "RMSE", "R2", "KGE", "FHV", "FLV"],
            "fill_nan": "no",
            "explainer": None,
            "rolling": None,
            "long_seq_pred": True,
            "calc_metrics": True,
        },
    }


def cmd(
    sub=None,
    source_cfgs=None,
    scaler=None,
    scaler_params=None,
    dataset=None,
    model_loader=None,
    sampler=None,
    fl_sample=None,
    fl_num_users=None,
    fl_local_ep=None,
    fl_local_bs=None,
    fl_frac=None,
    master_addr=None,
    port=None,
    ctx=None,
    rs=None,
    gage_id_file=None,
    gage_id=None,
    train_period=None,
    valid_period=None,
    test_period=None,
    opt=None,
    lr_scheduler=None,
    opt_param=None,
    batch_size=None,
    warmup_length=0,
    forecast_history=None,
    forecast_length=None,
    train_mode=None,
    train_epoch=None,
    save_epoch=None,
    save_iter=None,
    model_type=None,
    model_name=None,
    weight_path=None,
    continue_train=None,
    var_c=None,
    c_rm_nan=1,
    var_t=None,
    t_rm_nan=1,
    n_output=None,
    loss_func=None,
    model_hyperparam=None,
    dropout=None,
    weight_path_add=None,
    var_t_type=None,
    var_o=None,
    var_out=None,
    var_to_source_map=None,
    out_rm_nan=0,
    target_as_input=0,
    constant_only=0,
    gage_id_screen=None,
    loss_param=None,
    metrics=None,
    fill_nan=None,
    explainer=None,
    rolling=None,
    long_seq_pred=None,
    calc_metrics=None,
    start_epoch=1,
    stat_dict_file=None,
    num_workers=None,
    which_first_tensor=None,
    ensemble=0,
    ensemble_items=None,
    early_stopping=None,
    patience=None,
    min_time_unit=None,
    min_time_interval=None,
):
    """input args from cmd"""
    parser = argparse.ArgumentParser(
        description="Train a Time-Series Deep Learning Model for Basins"
    )
    parser.add_argument(
        "--sub", dest="sub", help="subset and sub experiment", default=sub, type=str
    )
    parser.add_argument(
        "--source_cfgs",
        dest="source_cfgs",
        help="configs for data sources",
        default=source_cfgs,
        type=json.loads,
    )
    parser.add_argument(
        "--scaler",
        dest="scaler",
        help="Choose a Scaler function",
        default=scaler,
        type=str,
    )
    parser.add_argument(
        "--scaler_params",
        dest="scaler_params",
        help="Parameters of the chosen Scaler function",
        default=scaler_params,
        type=json.loads,
    )
    parser.add_argument(
        "--dataset",
        dest="dataset",
        help="Choose a dataset class for PyTorch",
        default=dataset,
        type=str,
    )
    parser.add_argument(
        "--sampler",
        dest="sampler",
        help="None or KuaiSampler",
        default=sampler,
        type=str,
    )
    parser.add_argument(
        "--fl_sample",
        dest="fl_sample",
        help="sampling method for federated learning",
        default=fl_sample,
        type=str,
    )
    parser.add_argument(
        "--fl_num_users",
        dest="fl_num_users",
        help="number of users for federated learning",
        default=fl_num_users,
        type=int,
    )
    parser.add_argument(
        "--fl_local_ep",
        dest="fl_local_ep",
        help="number of local epochs for federated learning",
        default=fl_local_ep,
        type=int,
    )
    parser.add_argument(
        "--fl_local_bs",
        dest="fl_local_bs",
        help="local batch size for federated learning",
        default=fl_local_bs,
        type=float,
    )
    parser.add_argument(
        "--fl_frac",
        dest="fl_frac",
        help="the fraction of clients for federated learning",
        default=fl_frac,
        type=float,
    )
    parser.add_argument(
        "--master_addr",
        dest="master_addr",
        help="Running Context --default is localhost if you only train model on your computer",
        default=master_addr,
        nargs="+",
    )
    parser.add_argument(
        "--port",
        dest="port",
        help="Running Context -- which port do you want to use when using DistributedDataParallel",
        default=port,
        nargs="+",
    )
    parser.add_argument(
        "--ctx",
        dest="ctx",
        help="Running Context -- gpu num or cpu. E.g `--ctx 0 1` means run code in gpu 0 and 1; -1 means cpu",
        default=ctx,
        nargs="+",
    )
    parser.add_argument("--rs", dest="rs", help="random seed", default=rs, type=int)
    # There is something wrong with "bool", so I used 1 as True, 0 as False
    parser.add_argument(
        "--train_mode",
        dest="train_mode",
        help="If 0, no train or test, just read data; else train + test",
        default=train_mode,
        type=int,
    )
    parser.add_argument(
        "--train_epoch",
        dest="train_epoch",
        help="epoches of training period",
        default=train_epoch,
        type=int,
    )
    parser.add_argument(
        "--save_epoch",
        dest="save_epoch",
        help="save for every save_epoch epoches",
        default=save_epoch,
        type=int,
    )
    parser.add_argument(
        "--save_iter",
        dest="save_iter",
        help="save for every save_iter in save_epoches",
        default=save_iter,
        type=int,
    )
    parser.add_argument(
        "--loss_func",
        dest="loss_func",
        help="choose loss function",
        default=loss_func,
        type=str,
    )
    parser.add_argument(
        "--loss_param",
        dest="loss_param",
        help="choose parameters of loss function",
        default=loss_param,
        type=json.loads,
    )
    parser.add_argument(
        "--train_period",
        dest="train_period",
        help="The training period",
        default=train_period,
        nargs="+",
    )
    parser.add_argument(
        "--valid_period",
        dest="valid_period",
        help="The validating period",
        default=valid_period,
        nargs="+",
    )
    parser.add_argument(
        "--test_period",
        dest="test_period",
        help="The test period",
        default=test_period,
        nargs="+",
    )
    parser.add_argument(
        "--batch_size",
        dest="batch_size",
        help="batch_size",
        default=batch_size,
        type=int,
    )
    parser.add_argument(
        "--dropout",
        dest="dropout",
        help="dropout",
        default=dropout,
        type=float,
    )
    parser.add_argument(
        "--warmup_length",
        dest="warmup_length",
        help="Physical hydro models need warmup",
        default=warmup_length,
        type=int,
    )
    parser.add_argument(
        "--forecast_history",
        dest="forecast_history",
        help="length of time sequence when training in encoder part, for decoder-only models, forecast_history=0",
        default=forecast_history,
        type=int,
    )
    parser.add_argument(
        "--forecast_length",
        dest="forecast_length",
        help="length of time sequence when training in decoder part",
        default=forecast_length,
        type=int,
    )
    parser.add_argument(
        "--model_type",
        dest="model_type",
        help="The type of DL model",
        default=model_type,
        type=str,
    )
    parser.add_argument(
        "--model_name",
        dest="model_name",
        help="The name of DL model. now in the zoo",
        default=model_name,
        type=str,
    )
    parser.add_argument(
        "--weight_path",
        dest="weight_path",
        help="The weights of trained model",
        default=weight_path,
        type=str,
    )
    parser.add_argument(
        "--weight_path_add",
        dest="weight_path_add",
        help="More info about the weights of trained model",
        default=weight_path_add,
        type=json.loads,
    )
    parser.add_argument(
        "--continue_train",
        dest="continue_train",
        help="Continue to train the model from weight_path when continue_train>0",
        default=continue_train,
        type=int,
    )
    parser.add_argument(
        "--gage_id",
        dest="gage_id",
        help="just select some sites",
        default=gage_id,
        nargs="+",
    )
    parser.add_argument(
        "--gage_id_screen",
        dest="gage_id_screen",
        help="the criterion to chose some gages",
        default=gage_id_screen,
        type=json.loads,
    )
    parser.add_argument(
        "--gage_id_file",
        dest="gage_id_file",
        help="choose some sites from a file",
        default=gage_id_file,
        type=str,
    )
    parser.add_argument(
        "--opt", dest="opt", help="choose an optimizer", default=opt, type=str
    )
    parser.add_argument(
        "--opt_param",
        dest="opt_param",
        help="the optimizer parameters",
        default=opt_param,
        type=json.loads,
    )
    parser.add_argument(
        "--var_c", dest="var_c", help="types of attributes", default=var_c, nargs="+"
    )
    parser.add_argument(
        "--c_rm_nan",
        dest="c_rm_nan",
        help="if true, we remove NaN value for var_c data when scaling",
        default=c_rm_nan,
        type=int,
    )
    parser.add_argument(
        "--var_t", dest="var_t", help="types of forcing", default=var_t, nargs="+"
    )
    parser.add_argument(
        "--t_rm_nan",
        dest="t_rm_nan",
        help="if true, we remove NaN value for var_t data when scaling",
        default=t_rm_nan,
        type=int,
    )
    parser.add_argument(
        "--var_t_type",
        dest="var_t_type",
        help="types of forcing data_source",
        default=var_t_type,
        nargs="+",
    )
    parser.add_argument(
        "--var_o",
        dest="var_o",
        help="more other inputs except for var_c and var_t",
        default=var_o,
        type=json.loads,
    )
    parser.add_argument(
        "--var_out", dest="var_out", help="type of outputs", default=var_out, nargs="+"
    )
    parser.add_argument(
        "--var_to_source_map",
        dest="var_to_source_map",
        help="var_to_source_map",
        default=var_to_source_map,
        type=json.loads,
    )
    parser.add_argument(
        "--out_rm_nan",
        dest="out_rm_nan",
        help="if true, we remove NaN value for var_out data when scaling",
        default=out_rm_nan,
        type=int,
    )
    parser.add_argument(
        "--target_as_input",
        dest="target_as_input",
        help="if true, we will use target data as input for data assimilation or physics-based models",
        default=target_as_input,
        type=int,
    )
    parser.add_argument(
        "--constant_only",
        dest="constant_only",
        help="if true, we will only use attribute data as input for deep learning models; "
        "now it is only for dpl models and it is only used when target_as_input is False",
        default=constant_only,
        type=int,
    )
    parser.add_argument(
        "--n_output",
        dest="n_output",
        help="the number of output features",
        default=n_output,
        type=int,
    )
    parser.add_argument(
        "--model_hyperparam",
        dest="model_hyperparam",
        help="the model_hyperparam in model_cfgs",
        default=model_hyperparam,
        type=json.loads,
    )
    parser.add_argument(
        "--metrics",
        dest="metrics",
        help="The evaluating metrics",
        default=metrics,
        nargs="+",
    )
    parser.add_argument(
        "--fill_nan",
        dest="fill_nan",
        help="how to fill nan values when evaluating",
        default=fill_nan,
        nargs="+",
    )
    parser.add_argument(
        "--explainer",
        dest="explainer",
        help="explainer what when evaluating",
        default=explainer,
        nargs="+",
    )
    parser.add_argument(
        "--rolling",
        dest="rolling",
        help="if False, evaluate 1-period output with a rolling window",
        default=rolling,
        type=bool,
    )
    parser.add_argument(
        "--model_loader",
        dest="model_loader",
        help="the way to load weights of trained model",
        default=model_loader,
        type=int,
    )
    parser.add_argument(
        "--long_seq_pred",
        dest="long_seq_pred",
        help="if True, direct, one-step, long-term sequence prediction",
        default=long_seq_pred,
        type=bool,
    )
    parser.add_argument(
        "--calc_metrics",
        dest="calc_metrics",
        help="if False, calculate valid loss only",
        default=calc_metrics,
        type=bool,
    )
    parser.add_argument(
        "--start_epoch",
        dest="start_epoch",
        help="The index of epoch when starting training, default is 1. "
        "When retraining after an interrupt, it will be larger than 1",
        default=start_epoch,
        type=int,
    )
    parser.add_argument(
        "--stat_dict_file",
        dest="stat_dict_file",
        help="for testing sometimes such as pub cases, we need stat_dict_file from trained dataset",
        default=stat_dict_file,
        type=str,
    )
    parser.add_argument(
        "--num_workers",
        dest="num_workers",
        help="The number of workers used in Dataloader",
        default=num_workers,
        type=int,
    )
    parser.add_argument(
        "--which_first_tensor",
        dest="which_first_tensor",
        help="sequence_first or batch_first",
        default=which_first_tensor,
        type=str,
    )
    parser.add_argument(
        "--lr_scheduler",
        dest="lr_scheduler",
        help="The learning rate scheduler",
        default=lr_scheduler,
        type=json.loads,
    )
    parser.add_argument(
        "--ensemble",
        dest="ensemble",
        help="ensemble config",
        default=ensemble,
        type=int,
    )
    parser.add_argument(
        "--ensemble_items",
        dest="ensemble_items",
        help="ensemble config",
        default=ensemble_items,
        type=json.loads,
    )
    parser.add_argument(
        "--early_stopping",
        dest="early_stopping",
        help="early_stopping config",
        default=early_stopping,
        type=bool,
    )
    parser.add_argument(
        "--patience",
        dest="patience",
        help="patience config",
        default=patience,
        type=int,
    )
    parser.add_argument(
        "--min_time_unit",
        dest="min_time_unit",
        help="The min time type of the input data",
        default=min_time_unit,
        type=str,
    )
    parser.add_argument(
        "--min_time_interval",
        dest="min_time_interval",
        help="The min time interval of the input data",
        default=min_time_interval,
        type=int,
    )
    # To make pytest work in PyCharm, here we use the following code instead of "args = parser.parse_args()":
    # https://blog.csdn.net/u014742995/article/details/100119905
    args, unknown = parser.parse_known_args()
    return args


def update_nested_dict(d, keys, value):
    """update nested dict

    Parameters
    ----------
    d
        the dict to be updated
    keys
        the keys of the dict
    value
        the updated value of the dict
    """
    if len(keys) == 1:
        d[keys[0]] = value
    else:
        update_nested_dict(d[keys[0]], keys[1:], value)


def update_cfg(cfg_file, new_args):
    """
    Update default config with new arguments

    Parameters
    ----------
    cfg_file
        default config
    new_args
        new arguments

    Returns
    -------
    None
        in-place operation for cfg_file
    """
    print("update config file")
    project_dir = os.getcwd()
    result_dir = os.path.join(project_dir, "results")
    if os.path.exists(result_dir) is False:
        os.makedirs(result_dir)
    if new_args.sub is not None:
        subset, subexp = new_args.sub.split(os.sep)
        cfg_file["data_cfgs"]["validation_path"] = os.path.join(
            project_dir, "results", subset, subexp
        )
        cfg_file["data_cfgs"]["test_path"] = os.path.join(result_dir, subset, subexp)
    if new_args.source_cfgs is not None:
        cfg_file["data_cfgs"]["source_cfgs"] = new_args.source_cfgs
    if new_args.scaler is not None:
        cfg_file["data_cfgs"]["scaler"] = new_args.scaler
    if new_args.scaler_params is not None:
        cfg_file["data_cfgs"]["scaler_params"] = new_args.scaler_params
    if new_args.dataset is not None:
        cfg_file["data_cfgs"]["dataset"] = new_args.dataset
    if new_args.sampler is not None:
        cfg_file["data_cfgs"]["sampler"] = new_args.sampler
    if new_args.fl_sample is not None:
        if new_args.fl_sample not in ["basin", "region"]:
            # basin means each client is a basin
            raise ValueError("fl_sample must be 'basin' or 'region'")
        cfg_file["model_cfgs"]["fl_hyperparam"]["fl_sample"] = new_args.fl_sample
    if new_args.fl_num_users is not None:
        cfg_file["model_cfgs"]["fl_hyperparam"]["fl_num_users"] = new_args.fl_num_users
    if new_args.fl_local_ep is not None:
        cfg_file["model_cfgs"]["fl_hyperparam"]["fl_local_ep"] = new_args.fl_local_ep
    if new_args.fl_local_bs is not None:
        cfg_file["model_cfgs"]["fl_hyperparam"]["fl_local_bs"] = new_args.fl_local_bs
    if new_args.fl_frac is not None:
        cfg_file["model_cfgs1"]["fl_hyperparam"]["fl_frac"] = new_args.fl_frac
    if new_args.master_addr is not None:
        cfg_file["training_cfgs"]["master_addr"] = new_args.master_addr
    if new_args.port is not None:
        cfg_file["training_cfgs"]["port"] = new_args.port
    if new_args.ctx is not None:
        cfg_file["training_cfgs"]["device"] = new_args.ctx
    if new_args.rs is not None:
        cfg_file["training_cfgs"]["random_seed"] = new_args.rs
    if new_args.train_mode is not None:
        cfg_file["training_cfgs"]["train_mode"] = bool(new_args.train_mode > 0)
    if new_args.loss_func is not None:
        cfg_file["training_cfgs"]["criterion"] = new_args.loss_func
        if new_args.loss_param is not None:
            cfg_file["training_cfgs"]["criterion_params"] = new_args.loss_param
    if new_args.train_period is not None:
        cfg_file["data_cfgs"]["t_range_train"] = new_args.train_period
    if new_args.valid_period is not None:
        cfg_file["data_cfgs"]["t_range_valid"] = new_args.valid_period
    if new_args.test_period is not None:
        cfg_file["data_cfgs"]["t_range_test"] = new_args.test_period
    if (
        new_args.gage_id is not None
        and new_args.gage_id_file is not None
        or new_args.gage_id is None
        and new_args.gage_id_file is not None
    ):
        gage_id_lst = (
            pd.read_csv(new_args.gage_id_file, dtype={0: str}).iloc[:, 0].values
        )
        cfg_file["data_cfgs"]["object_ids"] = gage_id_lst.tolist()
    elif new_args.gage_id is not None:
        cfg_file["data_cfgs"]["object_ids"] = new_args.gage_id
    if new_args.opt is not None:
        cfg_file["training_cfgs"]["optimizer"] = new_args.opt
        if new_args.opt_param is not None:
            cfg_file["training_cfgs"]["optim_params"] = new_args.opt_param
        else:
            cfg_file["training_cfgs"]["optim_params"] = {}
    if new_args.var_c is not None:
        # I don't find a method to receive empty list for argparse, so if we input "None" or "" or " ", we treat it as []
        if (
            new_args.var_c == ["None"]
            or new_args.var_c == [""]
            or new_args.var_c == [" "]
        ):
            cfg_file["data_cfgs"]["constant_cols"] = []
        else:
            cfg_file["data_cfgs"]["constant_cols"] = new_args.var_c
    cfg_file["data_cfgs"]["constant_rm_nan"] = bool(new_args.c_rm_nan != 0)
    if new_args.var_t is not None:
        cfg_file["data_cfgs"]["relevant_cols"] = new_args.var_t
        print(
            "!!!!!!NOTE!!!!!!!!\n-------Please make sure the PRECIPITATION variable is in the 1st location in var_t setting!!---------"
        )
        print("If you have POTENTIAL_EVAPOTRANSPIRATION, please set it the 2nd!!!-")
    if new_args.var_t_type is not None:
        cfg_file["data_cfgs"]["relevant_types"] = new_args.var_t_type
    cfg_file["data_cfgs"]["relevant_rm_nan"] = bool(new_args.t_rm_nan != 0)
    if new_args.var_o is not None:
        cfg_file["data_cfgs"]["other_cols"] = new_args.var_o
    if new_args.var_out is not None:
        cfg_file["data_cfgs"]["target_cols"] = new_args.var_out
        print(
            "!!!!!!NOTE!!!!!!!!\n-------Please make sure the STREAMFLOW variable is in the 1st location in var_out setting!!---------"
        )
    if new_args.var_to_source_map is not None:
        cfg_file["data_cfgs"]["var_to_source_map"] = new_args.var_to_source_map
    cfg_file["data_cfgs"]["target_rm_nan"] = bool(new_args.out_rm_nan != 0)
    if new_args.target_as_input == 0:
        cfg_file["data_cfgs"]["target_as_input"] = False
        cfg_file["data_cfgs"]["constant_only"] = bool(new_args.constant_only != 0)
    else:
        cfg_file["data_cfgs"]["target_as_input"] = True
    if new_args.long_seq_pred is not None:
        cfg_file["evaluation_cfgs"]["long_seq_pred"] = new_args.long_seq_pred
    if new_args.calc_metrics is not None:
        cfg_file["evaluation_cfgs"]["calc_metrics"] = new_args.calc_metrics
    if new_args.train_epoch is not None:
        cfg_file["training_cfgs"]["epochs"] = new_args.train_epoch
    if new_args.save_epoch is not None:
        cfg_file["training_cfgs"]["save_epoch"] = new_args.save_epoch
    if new_args.save_iter is not None:
        cfg_file["training_cfgs"]["save_iter"] = new_args.save_iter
    if new_args.model_type is not None:
        cfg_file["model_cfgs"]["model_type"] = new_args.model_type
    if new_args.model_name is not None:
        cfg_file["model_cfgs"]["model_name"] = new_args.model_name
    if new_args.weight_path is not None:
        cfg_file["model_cfgs"]["weight_path"] = new_args.weight_path
        continue_train = bool(
            new_args.continue_train is not None and new_args.continue_train != 0
        )
        cfg_file["model_cfgs"]["continue_train"] = continue_train
    if new_args.weight_path_add is not None:
        cfg_file["model_cfgs"]["weight_path_add"] = new_args.weight_path_add
    if new_args.n_output is not None:
        cfg_file["training_cfgs"]["multi_targets"] = new_args.n_output
        if len(cfg_file["data_cfgs"]["target_cols"]) != new_args.n_output:
            raise AttributeError(
                "Please make sure size of vars in data_cfgs/target_cols is same as n_output"
            )
    if new_args.model_hyperparam is not None:
        # raise AttributeError("Please set the model_hyperparam!!!")
        cfg_file["model_cfgs"]["model_hyperparam"] = new_args.model_hyperparam
        if "batch_size" in new_args.model_hyperparam.keys():
            # TODO: batch_size's setting may conflict with batch_size's direct setting
            cfg_file["data_cfgs"]["batch_size"] = new_args.model_hyperparam[
                "batch_size"
            ]
            cfg_file["training_cfgs"]["batch_size"] = new_args.model_hyperparam[
                "batch_size"
            ]
        if "forecast_length" in new_args.model_hyperparam.keys():
            cfg_file["data_cfgs"]["forecast_length"] = new_args.model_hyperparam[
                "forecast_length"
            ]
        if "prec_window" in new_args.model_hyperparam.keys():
            cfg_file["data_cfgs"]["prec_window"] = new_args.model_hyperparam[
                "prec_window"
            ]
    if new_args.batch_size is not None:
        # raise AttributeError("Please set the batch_size!!!")
        batch_size = new_args.batch_size
        cfg_file["data_cfgs"]["batch_size"] = batch_size
        cfg_file["training_cfgs"]["batch_size"] = batch_size
    if new_args.min_time_unit is not None:
        if new_args.min_time_unit not in ["h", "D"]:
            raise ValueError("min_time_unit must be 'h' (HOURLY) or 'D' (DAILY)")
        cfg_file["data_cfgs"]["min_time_unit"] = new_args.min_time_unit
    if new_args.min_time_interval is not None:
        cfg_file["data_cfgs"]["min_time_interval"] = new_args.min_time_interval
    if new_args.metrics is not None:
        cfg_file["evaluation_cfgs"]["metrics"] = new_args.metrics
    if new_args.fill_nan is not None:
        cfg_file["evaluation_cfgs"]["fill_nan"] = new_args.fill_nan
    if new_args.explainer is not None:
        cfg_file["evaluation_cfgs"]["explainer"] = new_args.explainer
    if new_args.rolling is not None:
        cfg_file["evaluation_cfgs"]["rolling"] = new_args.rolling
    if new_args.model_loader is not None:
        cfg_file["evaluation_cfgs"]["model_loader"] = new_args.model_loader
    if new_args.warmup_length > 0:
        cfg_file["data_cfgs"]["warmup_length"] = new_args.warmup_length
        if "warmup_length" in new_args.model_hyperparam.keys() and (
            not cfg_file["data_cfgs"]["warmup_length"]
            == new_args.model_hyperparam["warmup_length"]
        ):
            raise RuntimeError(
                "Please set same warmup_length in model_cfgs and data_cfgs"
            )
    if new_args.forecast_history is not None:
        cfg_file["data_cfgs"]["forecast_history"] = new_args.forecast_history
    if new_args.forecast_length is not None:
        cfg_file["data_cfgs"]["forecast_length"] = new_args.forecast_length
    if new_args.start_epoch > 1:
        cfg_file["training_cfgs"]["start_epoch"] = new_args.start_epoch
    if new_args.stat_dict_file is not None:
        cfg_file["data_cfgs"]["stat_dict_file"] = new_args.stat_dict_file
    if new_args.num_workers is not None and new_args.num_workers > 0:
        cfg_file["training_cfgs"]["num_workers"] = new_args.num_workers
    if new_args.which_first_tensor is not None:
        cfg_file["training_cfgs"]["which_first_tensor"] = new_args.which_first_tensor
    if new_args.ensemble == 0:
        cfg_file["training_cfgs"]["ensemble"] = False
    else:
        cfg_file["training_cfgs"]["ensemble"] = True
    if new_args.ensemble_items is not None:
        cfg_file["training_cfgs"]["ensemble_items"] = new_args.ensemble_items
    if new_args.patience is not None:
        cfg_file["training_cfgs"]["patience"] = new_args.patience
    if new_args.early_stopping is not None:
        cfg_file["training_cfgs"]["early_stopping"] = new_args.early_stopping
    if new_args.lr_scheduler is not None:
        cfg_file["training_cfgs"]["lr_scheduler"] = new_args.lr_scheduler
    # print("the updated config:\n", json.dumps(cfg_file, indent=4, ensure_ascii=False))


def get_config_file(cfg_dir):
    json_files_lst = []
    json_files_ctime = []
    for file in os.listdir(cfg_dir):
        if (
            fnmatch.fnmatch(file, "*.json")
            and "_stat" not in file  # statistics json file
            and "_dict" not in file  # data cache json file
        ):
            json_files_lst.append(os.path.join(cfg_dir, file))
            json_files_ctime.append(os.path.getctime(os.path.join(cfg_dir, file)))
    sort_idx = np.argsort(json_files_ctime)
    cfg_file = json_files_lst[sort_idx[-1]]
    return hydro_file.unserialize_json(cfg_file)



"""
Author: Wenyu Ouyang
Date: 2023-10-05 16:16:48
LastEditTime: 2024-05-27 16:23:45
LastEditors: Wenyu Ouyang
Description: A test function for transfer learning
FilePath: \torchhydro\tests\test_pretrain_datafusion.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import train_and_evaluate
from torchhydro.models.seq2seq import DataEnhancedModel, GeneralSeq2Seq
import logging

logging.basicConfig(level=logging.INFO)
for logger_name in logging.root.manager.loggerDict:
    logger = logging.getLogger(logger_name)
    logger.setLevel(logging.INFO)


@pytest.fixture()
def config():
    weight_path = os.path.join(
        os.getcwd(), "results", "test_pretrain_enhanced", "ex1", "best_model.pth"
    )
    project_name = "test_pretrain_fusion/exp1"
    args = cmd(
        sub=project_name,
        source_cfgs={
            "source": "HydroMean",
            "source_path": {
                "forcing": "basins-origin/hour_data/1h/mean_data/mean_data_merged",
                "target": "basins-origin/hour_data/1h/mean_data/mean_data_merged",
                "attributes": "basins-origin/attributes.nc",
            },
        },
        ctx=[1],
        model_type="TransLearn",
        model_name="DataFusion",
        model_hyperparam={
            "original_model": DataEnhancedModel(
                original_model=GeneralSeq2Seq(
                    input_size=16,
                    output_size=1,
                    hidden_size=256,
                    forecast_length=24,
                    model_mode="dual",
                ),
                hidden_length=256,
            ),
            "cnn_size": 120,  # todo 有冗余
            "forecast_length": 24,
            "model_mode": "dual",
        },
        model_loader={"load_way": "best"},
        opt="Adam",
        lr_scheduler={
            "lr": 0.001,
            "lr_factor": 0.96,
        },
        loss_func="RMSESum",
        batch_size=256,
        forecast_history=168,
        forecast_length=24,
        train_period=[
            ("2022-07-08", "2022-09-29"),
        ],
        test_period=[
            ("2023-07-08", "2023-09-29"),
        ],
        valid_period=[
            ("2023-07-08", "2023-09-29"),
        ],
        dataset="MultiSourceDataset",
        sampler="HydroSampler",
        scaler="DapengScaler",
        weight_path=weight_path,
        weight_path_add={
            "freeze_params": [
                "encoder1.lstm.weight_ih_l0",
                "encoder1.lstm.weight_hh_l0",
                "encoder1.lstm.bias_ih_l0",
                "encoder1.lstm.bias_hh_l0",
                "encoder1.fc.weight",
                "encoder1.fc.bias",
                "encoder2.lstm.weight_ih_l0",
                "encoder2.lstm.weight_hh_l0",
                "encoder2.lstm.bias_ih_l0",
                "encoder2.lstm.bias_hh_l0",
                "encoder2.fc.weight",
                "encoder2.fc.bias",
                "decoder1.lstm.weight_ih_l0",
                "decoder1.lstm.weight_hh_l0",
                "decoder1.lstm.bias_ih_l0",
                "decoder1.lstm.bias_hh_l0",
                "decoder1.fc_out.weight",
                "decoder1.fc_out.bias",
                "decoder1.attention.attn.weight",
                "decoder2.lstm.weight_ih_l0",
                "decoder2.lstm.weight_hh_l0",
                "decoder2.lstm.bias_ih_l0",
                "decoder2.lstm.bias_hh_l0",
                "decoder2.fc_out.weight",
                "decoder2.fc_out.bias",
                "decoder2.attention.attn.weight",
                "lstm.weight_ih_l0",
                "lstm.weight_hh_l0",
                "lstm.bias_ih_l0",
                "lstm.bias_hh_l0",
                "fc.weight",
                "fc.bias",
            ]
        },
        continue_train=True,
        train_epoch=20,
        save_epoch=1,
        var_t=["gpm_tp(mm/h)", "smap(m3/m3)", "streamflow(m3/s)"],
        var_c=[
            "area",  # 面积
            "ele_mt_smn",  # 海拔(空间平均)
            "slp_dg_sav",  # 地形坡度 (空间平均)
            "sgr_dk_sav",  # 河流坡度 (平均)
            "for_pc_sse",  # 森林覆盖率
            "glc_cl_smj",  # 土地覆盖类型
            "run_mm_syr",  # 陆面径流 (流域径流的空间平均值)
            "inu_pc_slt",  # 淹没范围 (长期最大)
            "cmi_ix_syr",  # 气候湿度指数
            "aet_mm_syr",  # 实际蒸散发 (年平均)
            "snw_pc_syr",  # 雪盖范围 (年平均)
            "swc_pc_syr",  # 土壤水含量
            "gwt_cm_sav",  # 地下水位深度
            "cly_pc_sav",  # 土壤中的黏土、粉砂、砂粒含量
            "dor_pc_pva",  # 调节程度
        ],
        var_out=["streamflow(m3/s)"],
        gage_id=[
            "01181000",
            "01414500",
        ],
        which_first_tensor="batch",
        rolling=False,
        long_seq_pred=False,
        early_stopping=True,
        patience=4,
    )
    config_data = default_config_file()
    update_cfg(config_data, args)
    return config_data


def test_dataenhanced(config):
    train_and_evaluate(config)



"""
Author: Wenyu Ouyang
Date: 2024-04-17 12:32:26
LastEditTime: 2024-04-17 12:33:34
LastEditors: Xinzhuo Wu
Description:
FilePath: /torchhydro/torchhydro/models/seq2seq.py
Copyright (c) 2021-2024 Wenyu Ouyang. All rights reserved.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import random


class Attention(nn.Module):
    def __init__(self, hidden_dim):
        super(Attention, self).__init__()
        self.attn = nn.Linear(hidden_dim * 2, 1, bias=False)

    def forward(self, encoder_outputs, hidden):
        seq_len = encoder_outputs.shape[1]
        hidden = hidden.repeat(seq_len, 1, 1).transpose(0, 1)
        energy = torch.tanh(self.attn(torch.cat((hidden, encoder_outputs), dim=2)))
        return F.softmax(energy.squeeze(2), dim=1)


class AdditiveAttention(nn.Module):
    def __init__(self, hidden_dim):
        super(AdditiveAttention, self).__init__()
        self.W_q = nn.Linear(hidden_dim, hidden_dim, bias=False)
        self.W_k = nn.Linear(hidden_dim, hidden_dim, bias=False)
        self.v = nn.Linear(hidden_dim, 1, bias=False)

    def forward(self, encoder_outputs, hidden):
        seq_len = encoder_outputs.shape[1]
        hidden_transformed = self.W_q(hidden).repeat(seq_len, 1, 1).transpose(0, 1)
        encoder_outputs_transformed = self.W_k(encoder_outputs)
        combined = torch.tanh(hidden_transformed + encoder_outputs_transformed)
        scores = self.v(combined).squeeze(2)
        return F.softmax(scores, dim=1)


class DotProductAttention(nn.Module):
    def __init__(self, dropout=0.1):
        super(DotProductAttention, self).__init__()
        self.dropout = nn.Dropout(dropout)

    def forward(self, encoder_outputs, hidden):
        hidden_dim = encoder_outputs.shape[2]
        hidden_expanded = hidden.unsqueeze(1)
        scores = torch.bmm(
            hidden_expanded, encoder_outputs.transpose(1, 2)
        ) / math.sqrt(hidden_dim)
        attention_weights = F.softmax(scores, dim=-1)
        attention_weights = self.dropout(attention_weights)
        return attention_weights.squeeze(1)


class Encoder(nn.Module):
    def __init__(self, input_dim, hidden_dim, output_dim, num_layers=1, dropout=0.3):
        super(Encoder, self).__init__()
        self.hidden_dim = hidden_dim
        self.pre_fc = nn.Linear(input_dim, hidden_dim)
        self.pre_relu = nn.ReLU()
        self.lstm = nn.LSTM(hidden_dim, hidden_dim, num_layers, batch_first=True)
        self.dropout = nn.Dropout(dropout)
        self.fc = nn.Linear(hidden_dim, output_dim)

    def forward(self, x):
        x = self.pre_fc(x)
        x = self.pre_relu(x)
        outputs, (hidden, cell) = self.lstm(x)
        outputs = self.dropout(outputs)
        outputs = self.fc(outputs)
        return outputs, hidden, cell


class Decoder(nn.Module):
    def __init__(self, input_dim, output_dim, hidden_dim, num_layers=1, dropout=0.3):
        super(Decoder, self).__init__()
        self.hidden_dim = hidden_dim
        self.pre_fc = nn.Linear(input_dim, hidden_dim)
        self.pre_relu = nn.ReLU()
        self.lstm = nn.LSTM(hidden_dim, hidden_dim, num_layers, batch_first=True)
        self.dropout = nn.Dropout(dropout)
        self.fc_out = nn.Linear(hidden_dim, output_dim)

    def forward(self, input, hidden, cell):
        x = self.pre_fc(input)
        x = self.pre_relu(x)
        output, (hidden, cell) = self.lstm(x, (hidden, cell))
        output = self.dropout(output)
        output = self.fc_out(output)
        return output, hidden, cell


class StateTransferNetwork(nn.Module):
    def __init__(self, hidden_dim):
        super(StateTransferNetwork, self).__init__()
        self.fc_hidden = nn.Linear(hidden_dim, hidden_dim)
        self.fc_cell = nn.Linear(hidden_dim, hidden_dim)

    def forward(self, hidden, cell):
        transfer_hidden = torch.tanh(self.fc_hidden(hidden))
        transfer_cell = self.fc_cell(cell)
        return transfer_hidden, transfer_cell


class GeneralSeq2Seq(nn.Module):
    def __init__(
        self,
        en_input_size,
        de_input_size,
        output_size,
        hidden_size,
        forecast_length,
        prec_window=0,
        teacher_forcing_ratio=0.5,
    ):
        super(GeneralSeq2Seq, self).__init__()
        self.trg_len = forecast_length
        self.prec_window = prec_window
        self.teacher_forcing_ratio = teacher_forcing_ratio
        self.encoder = Encoder(
            input_dim=en_input_size, hidden_dim=hidden_size, output_dim=output_size
        )
        self.decoder = Decoder(
            input_dim=de_input_size, hidden_dim=hidden_size, output_dim=output_size
        )
        self.transfer = StateTransferNetwork(hidden_dim=hidden_size)

    def forward(self, *src):
        if len(src) == 3:
            src1, src2, trgs = src
        else:
            src1, src2 = src
            trgs = None
        encoder_outputs, hidden, cell = self.encoder(src1)
        hidden, cell = self.transfer(hidden, cell)
        outputs = []
        current_input = encoder_outputs[:, -1, :].unsqueeze(1)

        for t in range(self.trg_len):
            p = src2[:, t, :].unsqueeze(1)
            current_input = torch.cat((current_input, p), dim=2)
            output, hidden, cell = self.decoder(current_input, hidden, cell)
            outputs.append(output.squeeze(1))
            if trgs is None or self.teacher_forcing_ratio <= 0:
                current_input = output
            else:
                sm_trg = trgs[:, (self.prec_window + t), 1].unsqueeze(1).unsqueeze(1)
                if not torch.any(torch.isnan(sm_trg)).item():
                    use_teacher_forcing = random.random() < self.teacher_forcing_ratio
                    str_trg = output[:, :, 0].unsqueeze(2)
                    current_input = (
                        torch.cat((str_trg, sm_trg), dim=2)
                        if use_teacher_forcing
                        else output
                    )
                else:
                    current_input = output

        outputs = torch.stack(outputs, dim=1)
        prec_outputs = encoder_outputs[:, -self.prec_window, :].unsqueeze(1)
        outputs = torch.cat((prec_outputs, outputs), dim=1)
        return outputs


class DataEnhancedModel(GeneralSeq2Seq):
    def __init__(self, hidden_length, **kwargs):
        super(DataEnhancedModel, self).__init__(**kwargs)
        self.lstm = nn.LSTM(1, hidden_length, num_layers=1, batch_first=True)
        self.fc = nn.Linear(hidden_length, 6)

    def forward(self, *src):
        src1, src2, token = src
        processed_src1 = torch.unsqueeze(src1[:, :, 0], dim=2)
        out_src1, _ = self.lstm(processed_src1)
        out_src1 = self.fc(out_src1)
        combined_input = torch.cat((out_src1, src1[:, :, 1:]), dim=2)
        return super(DataEnhancedModel, self).forward(combined_input, src2, token)


class DataFusionModel(DataEnhancedModel):
    def __init__(self, input_dim, **kwargs):
        super(DataFusionModel, self).__init__(**kwargs)
        self.input_dim = input_dim

        self.fusion_layer = nn.Conv1d(
            in_channels=input_dim, out_channels=1, kernel_size=1
        )

    def forward(self, *src):
        src1, src2, token = src
        if self.input_dim == 3:
            processed_src1 = self.fusion_layer(
                src1[:, :, 0:3].permute(0, 2, 1)
            ).permute(0, 2, 1)
            combined_input = torch.cat((processed_src1, src1[:, :, 3:]), dim=2)
        else:
            processed_src1 = self.fusion_layer(
                src1[:, :, 0:2].permute(0, 2, 1)
            ).permute(0, 2, 1)
            combined_input = torch.cat((processed_src1, src1[:, :, 2:]), dim=2)

        return super(DataFusionModel, self).forward(combined_input, src2, token)


def gen_trg_mask(length, device):
    mask = torch.tril(torch.ones(length, length, device=device)) == 1

    mask = (
        mask.float()
        .masked_fill(mask == 0, float("-inf"))
        .masked_fill(mask == 1, float(0.0))
    )

    return mask


class Transformer(nn.Module):
    def __init__(
        self,
        n_encoder_inputs,
        n_decoder_inputs,
        n_decoder_output,
        channels=256,
        num_embeddings=512,
        nhead=8,
        num_layers=8,
        dropout=0.1,
        prec_window=0,
    ):
        super().__init__()

        self.input_pos_embedding = torch.nn.Embedding(num_embeddings, channels)
        self.target_pos_embedding = torch.nn.Embedding(num_embeddings, channels)

        encoder_layer = nn.TransformerEncoderLayer(
            d_model=channels,
            nhead=nhead,
            dropout=dropout,
            dim_feedforward=4 * channels,
        )
        decoder_layer = nn.TransformerDecoderLayer(
            d_model=channels,
            nhead=nhead,
            dropout=dropout,
            dim_feedforward=4 * channels,
        )

        self.encoder = torch.nn.TransformerEncoder(encoder_layer, num_layers)
        self.decoder = torch.nn.TransformerDecoder(decoder_layer, num_layers)

        self.input_projection = nn.Linear(n_encoder_inputs, channels)
        self.output_projection = nn.Linear(n_decoder_inputs, channels)

        self.linear = nn.Linear(channels, n_decoder_output)

        self.do = nn.Dropout(p=dropout)

    def encode_src(self, src):
        src_start = self.input_projection(src)

        in_sequence_len, batch_size = src_start.size(0), src_start.size(1)
        pos_encoder = (
            torch.arange(0, in_sequence_len, device=src.device)
            .unsqueeze(0)
            .repeat(batch_size, 1)
        )
        pos_encoder = self.input_pos_embedding(pos_encoder).permute(1, 0, 2)

        src = src_start + pos_encoder
        src = self.encoder(src) + src_start
        src = self.do(src)
        return src

    def decode_trg(self, trg, memory):
        trg_start = self.output_projection(trg)

        out_sequence_len, batch_size = trg_start.size(0), trg_start.size(1)
        pos_decoder = (
            torch.arange(0, out_sequence_len, device=trg.device)
            .unsqueeze(0)
            .repeat(batch_size, 1)
        )
        pos_decoder = self.target_pos_embedding(pos_decoder).permute(1, 0, 2)

        trg = pos_decoder + trg_start
        trg_mask = gen_trg_mask(out_sequence_len, trg.device)
        out = self.decoder(tgt=trg, memory=memory, tgt_mask=trg_mask) + trg_start
        out = self.do(out)
        out = self.linear(out)
        return out

    def forward(self, *x):
        src, trg = x
        src = self.encode_src(src)
        return self.decode_trg(trg=trg, memory=src)



"""
Author: Wenyu Ouyang
Date: 2023-10-19 21:34:29
LastEditTime: 2023-12-29 11:05:57
LastEditors: Xinzhuo Wu
Description: SHAP methods for deep learning models
FilePath: \torchhydro\torchhydro\explainers\shap.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
from typing import List
import torch
import numpy as np
import shap
import matplotlib.pyplot as plt
import seaborn as sns
import os


def plot_summary_shap_values(shap_values: torch.tensor, columns):
    mean_shap_values = shap_values.mean(dim=(0, 1))

    # Tensor to NumPy array to use matplotlib and seaborn
    mean_shap_values = mean_shap_values.abs().mean(dim=0).numpy()

    # seaborn barplot
    plt.figure(figsize=(10, 8))
    sns.barplot(x=mean_shap_values, y=columns, order=columns[::-1], orient="h")
    plt.xlabel("Mean Absolute SHAP Value")
    plt.title("Summary Plot of SHAP Values")
    plt.show()


def plot_summary_shap_values_over_time_series(shap_values: torch.tensor, columns):
    # Compute the absolute mean SHAP values over batches
    abs_mean_shap_values = shap_values.mean(dim=1).abs()

    # Further compute the mean over observations
    multi_shap_values = abs_mean_shap_values.mean(dim=1).numpy()

    # Plotting
    plt.figure(figsize=(10, 8))

    bottom = np.zeros(len(columns))
    for i, pred_shap_values in enumerate(multi_shap_values):
        plt.barh(columns[::-1], pred_shap_values, label=f"time-step {i}", left=bottom)
        bottom += pred_shap_values

    plt.xlabel("Mean Absolute SHAP Value Over Time Steps")
    plt.title("Summary Plot of SHAP Values Over Time Series")
    plt.legend()
    plt.show()


def jitter(values, jitter_strength=0.005):
    """Add some jitter to the values."""
    return values + jitter_strength * np.random.randn(*values.shape)


def plot_shap_values_from_history(shap_values: torch.tensor, history: torch.tensor):
    # Compute mean values across the specified axes
    mean_shap_values = shap_values.mean(dim=[0, 1])
    mean_history_values = history.mean(dim=1)
    figs: List[plt.Figure] = []
    # Loop through features and plot each scatter plot
    for feature_history, feature_shap_values in zip(
        mean_history_values.unbind(0), mean_shap_values.unbind(0)
    ):
        fig, ax = plt.subplots(figsize=(8, 6))
        # Convert tensor to numpy array for plotting
        feature_shap_values_np = feature_shap_values.numpy()
        feature_history_np = feature_history.numpy()
        scatter = ax.scatter(
            feature_shap_values_np,
            jitter(feature_shap_values_np),
            c=feature_history_np,
            cmap="bwr",  # Blue to Red color map
            marker="o",
        )
        ax.set_ylim(-0.05, 0.05)
        ax.set_xlabel("shap value")
        ax.set_yticks([])  # Hide y-axis ticks as they represent jittered values
        # Colorbar configuration
        cbar = plt.colorbar(scatter, ax=ax)
        cbar.set_label("feature values")
        figs.append(fig)
    return figs


def deep_explain_model_summary_plot(dl_model, test_dataset) -> None:
    """Generate feature summary plot for trained deep learning models

    Parameters
    ----------
    model
        trained model
    test_dataset
        test dataset
    """
    dl_model.eval()
    history = test_dataset.__getitem__(0)[0].unsqueeze(0)
    dl_model = dl_model.to("cpu")
    deep_explainer = shap.DeepExplainer(dl_model, history)
    shap_values = shap_results(deep_explainer, history)
    # summary plot shows overall feature ranking
    # by average absolute shap values
    fig = plot_summary_shap_values(shap_values, test_dataset.df.columns)
    abs_mean_shap_values = shap_values.mean(axis=["preds", "batches"])
    multi_shap_values = abs_mean_shap_values.mean(axis="observations")
    # summary plot for multi-step outputs
    # multi_shap_values = shap_values.apply_along_axis(np.mean, 'batches')
    fig = plot_summary_shap_values_over_time_series(
        shap_values, test_dataset.df.columns
    )
    history_numpy = torch.tensor(
        history.cpu().numpy(), names=["batches", "observations", "features"]
    )

    shap_values = shap_results(deep_explainer, history)
    figs = plot_shap_values_from_history(shap_values, history_numpy)


def shap_results(deep_explainer, history):
    result = deep_explainer.shap_values(history)
    result = np.stack(result)
    # shap_values needs to be 4-dimensional
    if len(result.shape) != 4:
        result = np.expand_dims(result, axis=0)
    result = torch.tensor(
        result, names=["preds", "batches", "observations", "features"]
    )

    return result


def plot_shap_value_heatmaps(shap_values: torch.tensor):
    # Compute the average shap values over batches
    average_shap_value_over_batches = shap_values.mean(dim=shap_values.dim() - 1)

    figs: List[plt.Figure] = []

    # Extracting the shapes of the data for the x and y ticks
    x = list(range(average_shap_value_over_batches.size(-1)))
    y = list(range(average_shap_value_over_batches.size(0)))

    # Iterate through features and plot each heatmap
    for shap_values_features in average_shap_value_over_batches.unbind(0):
        fig, ax = plt.subplots(figsize=(8, 6))

        # Convert tensor to numpy array for plotting
        shap_values_features_np = shap_values_features.numpy()

        # Plotting heatmap
        cax = ax.matshow(shap_values_features_np, cmap="bwr", aspect="auto")

        ax.set_xticks(x)
        ax.set_yticks(y)
        ax.set_xlabel("sequence history steps")
        ax.set_ylabel("prediction steps")

        # Adding colorbar to the right of the heatmap
        cbar = fig.colorbar(cax, ax=ax)
        cbar.set_label("feature values")

        figs.append(fig)

    return figs


def deep_explain_model_heatmap(dl_model, test_dataset) -> None:
    """Generate feature heatmap for prediction at a start time
    Parameters
    ----------
    model
        trained model
    test_dataset
        test dataset
    Returns
    -------
    None
    """
    dl_model.eval()
    history = test_dataset.__getitem__(0)[0]
    # background shape (L, N, M)
    # L - batch size, N - history length, M - feature size
    # for each element in each N x M batch in L,
    # attribute to each prediction in forecast len
    deep_explainer = shap.DeepExplainer(dl_model, history)
    shap_values = shap_results(deep_explainer, history)
    figs = plot_shap_value_heatmaps(shap_values)


def shap_summary_plot(dl_model, train_dataset, test_dataset) -> None:
    dl_model.eval()

    history_list = []
    for i in range(
        int(train_dataset.num_samples / len(train_dataset.data_cfgs["object_ids"]))
    ):
        h = train_dataset.__getitem__(i)[0]
        history_list.append(h)
    history = torch.cat(history_list).cuda()

    test_list = []
    for i in range(
        int(test_dataset.num_samples / len(test_dataset.data_cfgs["object_ids"]))
    ):
        t = test_dataset.__getitem__(i)[0]
        test_list.append(t)
    test = torch.cat(test_list).cuda()

    dl_model = dl_model.cuda()
    torch.backends.cudnn.enabled = False
    e = shap.DeepExplainer(dl_model, history)
    shap_values = e.shap_values(test)[0]

    shap_values_avg = (
        shap_values.squeeze(2)
        .reshape(shap_values.shape[0], shap_values.shape[1], -1)
        .mean(axis=-2)
    )
    test_tensor_avg = (
        test.squeeze(2).reshape(test.shape[0], test.shape[1], -1).mean(axis=-2)
    ).to("cpu")
    np.save(
        os.path.join(test_dataset.data_cfgs["test_path"], "shap_values_avg.npy"),
        shap_values_avg,
    )
    torch.save(
        test_tensor_avg,
        os.path.join(test_dataset.data_cfgs["test_path"], "test_tensor_avg.pth"),
    )
    shap.summary_plot(shap_values_avg, test_tensor_avg)
    plt.savefig(os.path.join(test_dataset.data_cfgs["test_path"], "shap.png"))



"""
Author: Wenyu Ouyang
Date: 2023-09-24 14:28:48
LastEditTime: 2023-12-18 09:11:55
LastEditors: Wenyu Ouyang
Description: A test for federated learning
FilePath: \torchhydro\tests\test_federated_learning.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import os
import pytest

from torchhydro import SETTING
from torchhydro.configs.config import cmd, default_config_file, update_cfg
from torchhydro.trainers.trainer import train_and_evaluate


@pytest.fixture()
def config():
    project_name = "test_camels/exp3"
    config_data = default_config_file()
    args = cmd(
        sub=project_name,
        source="CAMELS",
        source_path=os.path.join(
            SETTING["local_data_path"]["datasets-origin"], "camels", "camels_us"
        ),
        source_region="US",
        ctx=[-1],
        model_type="FedLearn",
        model_name="CpuLSTM",
        model_hyperparam={
            "n_input_features": 23,
            "n_output_features": 1,
            "n_hidden_states": 256,
        },
        gage_id=[
            "01013500",
            "01022500",
            "01030500",
            "01031500",
            "01047000",
            "01052500",
            "01054200",
            "01055000",
            "01057000",
            "01170100",
        ],
        batch_size=8,
        forecast_history=0,
        forecast_length=20,
        var_t=["dayl", "prcp", "srad", "tmax", "tmin", "vp"],
        # var_c=["None"],
        var_out=["streamflow"],
        dataset="StreamflowDataset",
        sampler="KuaiSampler",
        scaler="DapengScaler",
        train_epoch=2,
        save_epoch=1,
        te=2,
        train_period=["2000-10-01", "2001-10-01"],
        valid_period=["2001-10-01", "2002-10-01"],
        test_period=["2002-10-01", "2003-10-01"],
        loss_func="RMSESum",
        opt="Adam",
        # key is epoch, start from 1
        lr_scheduler={1: 1e-2, 2: 5e-3, 3: 1e-3},
        which_first_tensor="sequence",
    )
    update_cfg(config_data, args)
    return config_data


def test_train_evaluate(config):
    train_and_evaluate(config)



"""
Author: Wenyu Ouyang
Date: 2024-04-02 14:37:09
LastEditTime: 2024-07-10 09:26:07
LastEditors: Wenyu Ouyang
Description: A module for different data sources
FilePath: /torchhydro/torchhydro/datasets/data_sources.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import collections
import os
import numpy as np
import pandas as pd
import xarray as xr
import pint_xarray  # noqa but it is used in the code
from tqdm import tqdm

from hydroutils import hydro_time
from hydrodataset import Camels
from hydrodatasource.reader.data_source import SelfMadeHydroDataset


from torchhydro import CACHE_DIR, SETTING


class SupData4Camels:
    """A parent class for different data sources for CAMELS-US
    and also a class for reading streamflow data after 2014-12-31"""

    def __init__(self, supdata_dir=None) -> None:
        self.camels = Camels(
            data_path=os.path.join(
                SETTING["local_data_path"]["datasets-origin"], "camels", "camels_us"
            )
        )
        self.camels671_sites = self.camels.read_site_info()

        if supdata_dir is None:
            supdata_dir = os.path.join(
                SETTING["local_data_path"]["datasets-interim"],
                "camels_us",
            )
        self.data_source_dir = supdata_dir
        self.data_source_description = self.set_data_source_describe()

    def set_data_source_describe(self):
        return self.camels.data_source_description

    def read_ts_table(self, gage_id_lst=None, t_range=None, var_lst=None, **kwargs):
        """A parent function for reading camels timeseries data from csv or txt files.
        For different data sources, we need to implement this function.
        Here it is also a function for reading streamflow data after 2014-12-31.


        Parameters
        ----------
        gage_id_lst : _type_, optional
            basin ids, by default None
        t_range : _type_, optional
            time range, by default None
        var_lst : _type_, optional
            all variables including forcing and streamflow, by default None

        Raises
        ------
        NotImplementedError
            _description_
        """
        if gage_id_lst is None:
            gage_id_lst = self.all_basins
        if t_range is None:
            t_range = self.all_t_range
        if var_lst is None:
            var_lst = self.vars
        return self.camels.read_target_cols(
            gage_id_lst=gage_id_lst,
            t_range=t_range,
            target_cols=var_lst,
        )

    @property
    def all_basins(self):
        return self.camels671_sites["gauge_id"].values

    @property
    def all_t_range(self):
        # this is a left-closed right-open interval
        return ["1980-01-01", "2022-01-01"]

    @property
    def units(self):
        return ["ft^3/s"]

    @property
    def vars(self):
        return ["streamflow"]

    @property
    def ts_xrdataset_path(self):
        return CACHE_DIR.joinpath("camelsus_streamflow.nc")

    def cache_ts_xrdataset(self):
        """Save all timeseries data in a netcdf file in the cache directory"""
        basins = self.all_basins
        t_range = self.all_t_range
        times = hydro_time.t_range_days(t_range).tolist()
        variables = self.vars
        ts_data = self.read_ts_table(
            gage_id_lst=basins,
            t_range=t_range,
            var_lst=variables,
        )
        # All units' names are from Pint https://github.com/hgrecco/pint/blob/master/pint/default_en.txt
        units = self.units
        xr_data = xr.Dataset(
            data_vars={
                **{
                    variables[i]: (
                        ["basin", "time"],
                        ts_data[:, :, i],
                        {"units": units[i]},
                    )
                    for i in range(len(variables))
                }
            },
            coords={
                "basin": basins,
                "time": times,
            },
        )
        xr_data.to_netcdf(self.ts_xrdataset_path)

    def read_ts_xrdataset(self, gage_id_lst=None, t_range=None, var_lst=None):
        """Read all timeseries data from a netcdf file in the cache directory"""
        if not self.ts_xrdataset_path.exists():
            self.cache_ts_xrdataset()
        if var_lst is None:
            return None
        ts = xr.open_dataset(self.ts_xrdataset_path)
        all_vars = ts.data_vars
        if any(var not in ts.variables for var in var_lst):
            raise ValueError(f"var_lst must all be in {all_vars}")
        return ts[var_lst].sel(basin=gage_id_lst, time=slice(t_range[0], t_range[1]))

    def read_attr_xrdataset(self, gage_id_lst=None, var_lst=None):
        return self.camels.read_attr_xrdataset(gage_id_lst=gage_id_lst, var_lst=var_lst)


# the following is a dict for different data sources
class ModisEt4Camels(SupData4Camels):
    """
    A datasource class for MODIS ET data of basins in CAMELS.

    Attributes data come from CAMELS.
    ET data include:
        PMLV2 (https://doi.org/10.1016/j.rse.2018.12.031)
        MODIS16A2v006 (https://developers.google.com/earth-engine/datasets/catalog/MODIS_006_MOD16A2?hl=en#bands)
        MODIS16A2v105 (https://developers.google.com/earth-engine/datasets/catalog/MODIS_NTSG_MOD16A2_105?hl=en#description)
    """

    def __init__(self, supdata_dir=None):
        """
        Initialize a ModisEt4Camels instance.

        Parameters
        ----------
        supdata_dir
            a list including the data file directory for the instance and CAMELS's path

        """
        if supdata_dir is None:
            supdata_dir = os.path.join(
                SETTING["local_data_path"]["datasets-interim"],
                "camels_us",
                "modiset4camels",
            )
        super().__init__(supdata_dir)

    @property
    def all_t_range(self):
        # this is a left-closed right-open interval
        return ["2001-01-01", "2022-01-01"]

    @property
    def units(self):
        return [
            "gC/m^2/d",
            "mm/day",
            "mm/day",
            "mm/day",
            "mm/day",
            "mm/day",
            "mm/day",
            "mm/day",
            "mm/day",
            "mm/day",
            "dimensionless",
        ]

    @property
    def vars(self):
        return [
            # PMLV2
            "GPP",
            "Ec",
            "Es",
            "Ei",
            "ET_water",
            # PML_V2's ET = Ec + Es + Ei
            "ET_sum",
            # MODIS16A2
            "ET",
            "LE",
            "PET",
            "PLE",
            "ET_QC",
        ]

    @property
    def ts_xrdataset_path(self):
        return CACHE_DIR.joinpath("camelsus_modiset.nc")

    def set_data_source_describe(self):
        et_db = self.data_source_dir
        # ET
        et_basin_mean_dir = os.path.join(et_db, "basin_mean_forcing")
        modisa16v105_dir = os.path.join(et_basin_mean_dir, "MOD16A2_105_CAMELS")
        modisa16v006_dir = os.path.join(et_basin_mean_dir, "MOD16A2_006_CAMELS")
        pmlv2_dir = os.path.join(et_basin_mean_dir, "PML_V2_CAMELS")
        if not os.path.isdir(et_basin_mean_dir):
            raise NotADirectoryError(
                "Please check if you have downloaded the data and put it in the correct dir"
            )
        return collections.OrderedDict(
            MODIS_ET_CAMELS_DIR=et_db,
            MODIS_ET_CAMELS_MEAN_DIR=et_basin_mean_dir,
            MOD16A2_CAMELS_DIR=modisa16v006_dir,
            PMLV2_CAMELS_DIR=pmlv2_dir,
        )

    def read_ts_table(
        self,
        gage_id_lst=None,
        t_range=None,
        var_lst=None,
        reduce_way="mean",
        **kwargs,
    ):
        """
        Read ET data.

        Parameters
        ----------
        gage_id_lst
            the ids of gages in CAMELS
        t_range
            the start and end periods
        target_cols
            the forcing var types
        reduce_way
            how to do "reduce" -- mean or sum; the default is "mean"

        Returns
        -------
        np.array
            return an np.array
        """

        assert len(t_range) == 2
        assert all(x < y for x, y in zip(gage_id_lst, gage_id_lst[1:]))
        # Data is not daily. For convenience, we fill NaN values in gap periods.
        # For example, the data is in period 1 (1-8 days), then there is one data in the 1st day while the rest are NaN
        t_range_list = hydro_time.t_range_days(t_range)
        nt = t_range_list.shape[0]
        x = np.empty([len(gage_id_lst), nt, len(var_lst)])
        for k in tqdm(range(len(gage_id_lst)), desc="Read MODIS ET data for CAMELS-US"):
            # two way to read data are provided:
            # 1. directly read data: the data is sum of 8 days
            # 2. calculate daily mean value of 8 days
            data = self.read_basin_mean_modiset(
                gage_id_lst[k], var_lst, t_range_list, reduce_way=reduce_way
            )
            x[k, :, :] = data
        return x

    def read_basin_mean_modiset(
        self, usgs_id, var_lst, t_range_list, reduce_way
    ) -> np.array:
        """
        Read modis ET from PMLV2 and MOD16A2

        Parameters
        ----------
        usgs_id
            ids of basins
        var_lst
            et variables from PMLV2 or/and MOD16A2
        t_range_list
            daily datetime list
        reduce_way
            how to do "reduce" -- mean or sum; the default is "sum"

        Returns
        -------
        np.array
            ET data
        """
        gage_id_df = self.camels671_sites
        huc = gage_id_df[gage_id_df["gauge_id"] == usgs_id]["huc_02"].values[0]

        modis16a2_data_folder = self.data_source_description["MOD16A2_CAMELS_DIR"]
        pmlv2_data_folder = self.data_source_description["PMLV2_CAMELS_DIR"]
        pmlv2_data_file = os.path.join(
            pmlv2_data_folder, huc, f"{usgs_id}_lump_pmlv2_et.txt"
        )
        modis16a2_data_file = os.path.join(
            modis16a2_data_folder, huc, f"{usgs_id}_lump_modis16a2v006_et.txt"
        )
        pmlv2_data_temp = pd.read_csv(pmlv2_data_file, header=None, skiprows=1)
        modis16a2_data_temp = pd.read_csv(modis16a2_data_file, header=None, skiprows=1)
        pmlv2_lst = [
            "Year",
            "Mnth",
            "Day",
            "Hr",
            "GPP",
            "Ec",
            "Es",
            "Ei",
            "ET_water",
            "ET_sum",
        ]  # PMLV2
        modis16a2_lst = [
            "Year",
            "Mnth",
            "Day",
            "Hr",
            "ET",
            "LE",
            "PET",
            "PLE",
            "ET_QC",
        ]  # MODIS16A2
        df_date_pmlv2 = pmlv2_data_temp[[0, 1, 2]]
        df_date_pmlv2.columns = ["year", "month", "day"]
        df_date_modis16a2 = modis16a2_data_temp[[0, 1, 2]]
        df_date_modis16a2.columns = ["year", "month", "day"]
        ind1_pmlv2, ind2_pmlv2, t_range_final_pmlv2 = self.date_intersect(
            df_date_pmlv2, t_range_list
        )
        (
            ind1_modis16a2,
            ind2_modis16a2,
            t_range_final_modis16a2,
        ) = self.date_intersect(df_date_modis16a2, t_range_list)

        nf = len(var_lst)
        nt = len(t_range_list)
        out = np.full([nt, nf], np.nan)

        for k in range(nf):
            if var_lst[k] in pmlv2_lst:
                if len(t_range_final_pmlv2) == 0:
                    # no data, just skip this var
                    continue
                if var_lst[k] == "ET_sum":
                    # No such item in original PML_V2 data
                    et_3components = self.read_basin_mean_modiset(
                        usgs_id, ["Ec", "Es", "Ei"], t_range_list, reduce_way
                    )
                    # it si equal to sum of 3 components
                    out[:, k] = np.sum(et_3components, axis=-1)
                    continue
                ind = pmlv2_lst.index(var_lst[k])
                if reduce_way == "sum":
                    out[ind2_pmlv2, k] = pmlv2_data_temp[ind].values[ind1_pmlv2]
                elif reduce_way == "mean":
                    days_interval = [y - x for x, y in zip(ind2_pmlv2, ind2_pmlv2[1:])]
                    if (
                        t_range_final_pmlv2[-1].item().month == 12
                        and t_range_final_pmlv2[-1].item().day == 31
                    ):
                        final_timedelta = (
                            t_range_final_pmlv2[-1].item()
                            - t_range_final_pmlv2[ind2_pmlv2[-1]].item()
                        )
                        final_day_interval = [final_timedelta.days]
                    else:
                        final_day_interval = [8]
                    days_interval = np.array(days_interval + final_day_interval)
                    # there may be some missing data, so that some interval will be larger than 8
                    days_interval[np.where(days_interval > 8)] = 8
                    out[ind2_pmlv2, k] = (
                        pmlv2_data_temp[ind].values[ind1_pmlv2] / days_interval
                    )
                else:
                    raise NotImplementedError("We don't have such a reduce way")
            elif var_lst[k] in modis16a2_lst:
                if len(t_range_final_modis16a2) == 0:
                    # no data, just skip this var
                    continue
                ind = modis16a2_lst.index(var_lst[k])
                if reduce_way == "sum":
                    out[ind2_modis16a2, k] = modis16a2_data_temp[ind].values[
                        ind1_modis16a2
                    ]
                elif reduce_way == "mean":
                    days_interval = [
                        y - x for x, y in zip(ind2_modis16a2, ind2_modis16a2[1:])
                    ]
                    if (
                        t_range_final_modis16a2[-1].item().month == 12
                        and t_range_final_modis16a2[-1].item().day == 31
                    ):
                        final_timedelta = (
                            t_range_final_modis16a2[-1].item()
                            - t_range_final_modis16a2[ind2_modis16a2[-1]].item()
                        )
                        final_day_interval = [final_timedelta.days]
                    else:
                        final_day_interval = [8]
                    days_interval = np.array(days_interval + final_day_interval)
                    # there may be some missing data, so that some interval will be larger than 8
                    days_interval[np.where(days_interval > 8)] = 8
                    out[ind2_modis16a2, k] = (
                        modis16a2_data_temp[ind].values[ind1_modis16a2] / days_interval
                    )
                else:
                    raise NotImplementedError("We don't have such a reduce way")
            else:
                raise NotImplementedError("No such var type now")
        # unit is 0.1mm/day(or 8/5/6days), so multiply it with 0.1 to transform to mm/day(or 8/5/6days))
        # TODO: only valid for MODIS, for PMLV2, we need to check the unit
        out = out * 0.1
        return out

    @staticmethod
    def date_intersect(df_date, t_range_list):
        date = pd.to_datetime(df_date).values.astype("datetime64[D]")
        if (
            np.datetime64(f"{str(date[-1].astype(object).year)}-12-31")
            > date[-1]
            > np.datetime64(f"{str(date[-1].astype(object).year)}-12-24")
        ):
            final_date = np.datetime64(f"{str(date[-1].astype(object).year + 1)}-01-01")
        else:
            final_date = date[-1] + np.timedelta64(8, "D")
        date_all = hydro_time.t_range_days(
            hydro_time.t_days_lst2range([date[0], final_date])
        )
        t_range_final = np.intersect1d(date_all, t_range_list)
        [c, ind1, ind2] = np.intersect1d(date, t_range_final, return_indices=True)
        return ind1, ind2, t_range_final


class Nldas4Camels(SupData4Camels):
    """
    A datasource class for geo attributes data, NLDAS v2 forcing data, and streamflow data of basins in CAMELS.

    The forcing data are basin mean values. Attributes and streamflow data come from CAMELS.
    """

    def __init__(self, supdata_dir=None):
        """
        Initialize a Nldas4Camels instance.

        Parameters
        ----------
        supdata_dir
            a list including the data file directory for the instance and CAMELS's path

        """
        if supdata_dir is None:
            supdata_dir = os.path.join(
                SETTING["local_data_path"]["datasets-interim"],
                "camels_us",
                "nldas4camels",
            )
        super().__init__(supdata_dir=supdata_dir)

    @property
    def units(self):
        return [
            "°C",
            "dimensionless",
            "Pa",
            "m/s",
            "m/s",
            "W/m^2",
            "dimensionless",
            "W/m^2",
            "J/kg",
            # unit of potential_evaporation and total_precipitation is kg/m^2 (for a day),
            # we use rho=1000kg/m^3 as water's density to transform these two variables‘ unit to mm/day
            # so it's 10^-3 m /day and it is just mm/day, hence we don't need to transform actually
            "mm/day",
            "mm/day",
        ]

    @property
    def vars(self):
        return [
            "temperature",
            "specific_humidity",
            "pressure",
            "wind_u",
            "wind_v",
            "longwave_radiation",
            "convective_fraction",
            "shortwave_radiation",
            "potential_energy",
            "potential_evaporation",
            "total_precipitation",
        ]

    @property
    def ts_xrdataset_path(self):
        return CACHE_DIR.joinpath("camelsus_nldas.nc")

    def set_data_source_describe(self):
        nldas_db = self.data_source_dir
        # forcing
        nldas_forcing_basin_mean_dir = os.path.join(nldas_db, "basin_mean_forcing")
        if not os.path.isdir(nldas_forcing_basin_mean_dir):
            raise NotADirectoryError(
                "Please check if you have downloaded the data and put it in the correct dir"
            )
        return collections.OrderedDict(
            NLDAS_CAMELS_DIR=nldas_db,
            NLDAS_CAMELS_MEAN_DIR=nldas_forcing_basin_mean_dir,
        )

    def read_basin_mean_nldas(self, usgs_id, var_lst, t_range_list):
        gage_id_df = self.camels671_sites
        huc = gage_id_df[gage_id_df["gauge_id"] == usgs_id]["huc_02"].values[0]

        data_folder = self.data_source_description["NLDAS_CAMELS_MEAN_DIR"]
        data_file = os.path.join(
            data_folder, huc, f"{usgs_id}_lump_nldas_forcing_leap.txt"
        )
        data_temp = pd.read_csv(data_file, sep=r"\s+", header=None, skiprows=1)
        forcing_lst = ["Year", "Mnth", "Day", "Hr"] + self.vars
        df_date = data_temp[[0, 1, 2]]
        df_date.columns = ["year", "month", "day"]
        date = pd.to_datetime(df_date).values.astype("datetime64[D]")

        nf = len(var_lst)
        [c, ind1, ind2] = np.intersect1d(date, t_range_list, return_indices=True)
        nt = c.shape[0]
        out = np.empty([nt, nf])

        for k in range(nf):
            ind = forcing_lst.index(var_lst[k])
            if "potential_evaporation" in var_lst[k]:
                pet = data_temp[ind].values
                # there are a few negative values for pet, set them 0
                pet[pet < 0] = 0.0
                out[ind2, k] = pet[ind1]
            else:
                out[ind2, k] = data_temp[ind].values[ind1]
        return out

    def read_ts_table(
        self,
        gage_id_lst: list = None,
        t_range: list = None,
        var_lst: list = None,
        **kwargs,
    ) -> np.array:
        """
        Read forcing data.

        Parameters
        ----------
        gage_id_lst
            the ids of gages in CAMELS
        t_range
            the start and end periods
        var_lst
            the forcing var types

        Returns
        -------
        np.array
            return an np.array
        """

        assert len(t_range) == 2
        assert all(x < y for x, y in zip(gage_id_lst, gage_id_lst[1:]))

        t_range_list = hydro_time.t_range_days(t_range)
        nt = t_range_list.shape[0]
        x = np.empty([len(gage_id_lst), nt, len(var_lst)])
        for k in tqdm(
            range(len(gage_id_lst)), desc="Read NLDAS forcing data for CAMELS-US"
        ):
            data = self.read_basin_mean_nldas(gage_id_lst[k], var_lst, t_range_list)
            x[k, :, :] = data
        return x


class Smap4Camels(SupData4Camels):
    """
    A datasource class for geo attributes data, forcing data, and SMAP data of basins in CAMELS.
    """

    def __init__(self, supdata_dir=None):
        """
        Parameters
        ----------
        supdata_dir
            a list including the data file directory for the instance and CAMELS's path

        """
        if supdata_dir is None:
            supdata_dir = os.path.join(
                SETTING["local_data_path"]["datasets-interim"],
                "camels_us",
                "smap4camels",
            )
        super().__init__(supdata_dir=supdata_dir)

    @property
    def all_t_range(self):
        # this is a left-closed right-open interval
        return ["2015-04-01", "2021-10-04"]

    @property
    def units(self):
        return ["mm", "mm", "dimensionless", "dimensionless", "dimensionless"]

    @property
    def vars(self):
        return ["ssm", "susm", "smp", "ssma", "susma"]

    @property
    def ts_xrdataset_path(self):
        return CACHE_DIR.joinpath("camelsus_smap.nc")

    def set_data_source_describe(self):
        # forcing
        smap_db = self.data_source_dir
        if not os.path.isdir(smap_db):
            raise NotADirectoryError(
                "Please check if you have downloaded the data and put it in the correct dir"
            )
        smap_data_dir = os.path.join(smap_db, "NASA_USDA_SMAP_CAMELS")
        return collections.OrderedDict(
            SMAP_CAMELS_DIR=smap_db, SMAP_CAMELS_MEAN_DIR=smap_data_dir
        )

    def read_ts_table(self, gage_id_lst=None, t_range=None, var_lst=None, **kwargs):
        """
        Read SMAP basin mean data

        More detials about NASA-USDA Enhanced SMAP data could be seen in:
        https://explorer.earthengine.google.com/#detail/NASA_USDA%2FHSL%2FSMAP10KM_soil_moisture

        Parameters
        ----------
        gage_id_lst
            the ids of gages in CAMELS
        t_range
            the start and end periods
        target_cols
            the var types

        Returns
        -------
        np.array
            return an np.array
        """
        # Data is not daily. For convenience, we fill NaN values in gap periods.
        # For example, the data is in period 1 (1-3 days), then there is one data in the 1st day while the rest are NaN
        t_range_list = hydro_time.t_range_days(t_range)
        nt = t_range_list.shape[0]
        x = np.empty([len(gage_id_lst), nt, len(var_lst)])
        for k in tqdm(
            range(len(gage_id_lst)), desc="Read NSDA-SMAP data for CAMELS-US"
        ):
            # two way to read data are provided:
            # 1. directly read data: the data is sum of 8 days
            # 2. calculate daily mean value of 8 days
            data = self.read_basin_mean_smap(gage_id_lst[k], var_lst, t_range_list)
            x[k, :, :] = data
        return x

    def read_basin_mean_smap(self, usgs_id, var_lst, t_range_list):
        gage_id_df = self.camels671_sites
        huc = gage_id_df[gage_id_df["gauge_id"] == usgs_id]["huc_02"].values[0]

        data_folder = self.data_source_description["SMAP_CAMELS_MEAN_DIR"]
        data_file = os.path.join(data_folder, huc, f"{usgs_id}_lump_nasa_usda_smap.txt")
        data_temp = pd.read_csv(data_file, sep=",", header=None, skiprows=1)
        smap_var_lst = [
            "Year",
            "Mnth",
            "Day",
            "Hr",
            "ssm",
            "susm",
            "smp",
            "ssma",
            "susma",
        ]
        df_date = data_temp[[0, 1, 2]]
        df_date.columns = ["year", "month", "day"]
        date = pd.to_datetime(df_date).values.astype("datetime64[D]")
        [c, ind1, ind2] = np.intersect1d(date, t_range_list, return_indices=True)

        nf = len(var_lst)
        nt = len(t_range_list)
        out = np.full([nt, nf], np.nan)

        for k in range(nf):
            ind = smap_var_lst.index(var_lst[k])
            out[ind2, k] = data_temp[ind].values[ind1]
        return out


data_sources_dict = {
    "camels_us": Camels,
    "selfmadehydrodataset": SelfMadeHydroDataset,
    "usgs4camels": SupData4Camels,
    "modiset4camels": ModisEt4Camels,
    "nldas4camels": Nldas4Camels,
    "smap4camels": Smap4Camels,
}



"""
Author: Wenyu Ouyang
Date: 2021-12-31 11:08:29
LastEditTime: 2024-09-10 10:56:55
LastEditors: Wenyu Ouyang
Description: Dicts including models (which are seq-first), losses, and optims
FilePath: \torchhydro\torchhydro\models\model_dict_function.py
Copyright (c) 2021-2022 Wenyu Ouyang. All rights reserved.
"""

from torchhydro.models.cudnnlstm import (
    CudnnLstmModel,
    LinearCudnnLstmModel,
    CNN1dLCmodel,
    CudnnLstmModelLstmKernel,
    CudnnLstmModelMultiOutput,
    CpuLstmModel,
)

from torchhydro.models.simple_lstm import SimpleLSTMForecast
from torchhydro.models.seq2seq import (
    GeneralSeq2Seq,
    DataEnhancedModel,
    DataFusionModel,
    Transformer,
)
from torch.optim import Adam, SGD, Adadelta
from torchhydro.models.crits import (
    RMSELoss,
    RmseLoss,
    MultiOutLoss,
    UncertaintyWeights,
    DynamicTaskPrior,
    MultiOutWaterBalanceLoss,
    NSELoss,
    MAPELoss,
    MASELoss,
    MAELoss,
    QuantileLoss,
)
from torchhydro.models.dpl4xaj import DplLstmXaj, DplAnnXaj
from torchhydro.models.dpl4xaj_nn4et import DplLstmNnModuleXaj
from torchhydro.models.spplstm import SPP_LSTM_Model, SPP_LSTM_Model_2

"""
Utility dictionaries to map a string to a class.
"""
pytorch_model_dict = {
    "KuaiLSTM": CudnnLstmModel,
    "CpuLSTM": CpuLstmModel,
    "KaiLSTM": LinearCudnnLstmModel,
    "DapengCNNLSTM": CNN1dLCmodel,
    "LSTMKernel": CudnnLstmModelLstmKernel,
    "KuaiLSTMMultiOut": CudnnLstmModelMultiOutput,
    "DplLstmXaj": DplLstmXaj,
    "DplAttrXaj": DplAnnXaj,
    "SPPLSTM": SPP_LSTM_Model,
    "SimpleLSTMForecast": SimpleLSTMForecast,
    "SPPLSTM2": SPP_LSTM_Model_2,
    "Seq2Seq": GeneralSeq2Seq,
    "DataEnhanced": DataEnhancedModel,
    "DataFusion": DataFusionModel,
    "Transformer": Transformer,
    "DplNnModuleXaj": DplLstmNnModuleXaj,
}

pytorch_criterion_dict = {
    "RMSE": RMSELoss,
    # xxxSum means that calculate the criterion for each "feature"(the final dim of output), then sum them up
    "RMSESum": RmseLoss,
    "MultiOutLoss": MultiOutLoss,
    "UncertaintyWeights": UncertaintyWeights,
    "DynamicTaskPrior": DynamicTaskPrior,
    "MultiOutWaterBalanceLoss": MultiOutWaterBalanceLoss,
    "NSELoss": NSELoss,
    "MAPELoss": MAPELoss,
    "MASELoss": MASELoss,
    "MAELoss": MAELoss,
    "QuantileLoss": QuantileLoss,
}

pytorch_opt_dict = {"Adam": Adam, "SGD": SGD, "Adadelta": Adadelta}
